from fastapi import APIRouter, HTTPException, Query
from brave.api.config.db import get_engine
from brave.microbe.models.core import t_taxonomy
from brave.microbe.service import study_service
from brave.microbe.service import  disease_service
from brave.microbe.service import  chemicals_and_drugs_service
from brave.microbe.service import  diet_and_food_service
from brave.microbe.service import  association_service
from brave.microbe.service import  mesh_service

from sqlalchemy import insert
import brave.microbe.service.taxonomy_service as taxonomy_service
from brave.microbe.schemas.entity import AddMeshNode, DetailsNodeQuery, NodeQuery, PageEntity
from brave.microbe.service import graph_service
from brave.api.config.config import  get_graph
import json
entity_api = APIRouter(prefix="/entity")


@entity_api.get("/init/{entity}")
def import_taxonomy(entity,category=None):
    with get_engine().begin() as conn:
        if entity == "taxonomy":
            return  taxonomy_service.init(conn)
        elif entity == "disease":
            return disease_service.init(conn)
        elif entity == "mesh":
            return mesh_service.init(conn,category)
        else:
            raise ValueError("Unsupported entity type")
        # 读取三张表并合并

def formt_items_json(item):
    if "tags" in item and item["tags"]:
        try:
            item["tags"] = json.loads(item["tags"])
        except:
            pass
    if "short_name" in item and item["short_name"]:
        try:
            item["short_name"] = json.loads(item["short_name"])
        except:
            pass
    return item
def formt_items(item):
    item = dict(item)
    return formt_items_json(item)
    

@entity_api.post("/page/{entity}")
async def page_entity(entity: str, query: PageEntity):
    with get_engine().begin() as conn:
        if entity == "taxonomy":    
            result = taxonomy_service.page_taxonomy_v3(conn, query)
        elif entity == "study":
            result = study_service.page_study(conn, query)  
        elif entity == "disease":
            result = disease_service.page_disease(conn, query)
        elif entity == "chemicals_and_drugs":
            result = chemicals_and_drugs_service.page(conn, query)
        elif entity == "diet_and_food":
            result = diet_and_food_service.page(conn, query)
        elif entity == "association":
            result = association_service.page(conn, query)
        elif entity =="mesh":
            result = mesh_service.page(conn,query)
        else:
            raise ValueError("Unsupported entity type")
    
    entity_ids = [item["entity_id"] for item in result["items"]]
    entity_ids_map = graph_service.check_nodes_exist_batch(entity_ids,entity)
    result["items"] = [{"is_exist_graph":entity_ids_map[item["entity_id"]],**formt_items(item)} for item in result["items"]]
    return result

@entity_api.get("/get/{entity}/{entity_id}")
async def get_entity(entity: str, entity_id: str):
    with get_engine().begin() as conn:
        if entity=="taxonomy":  
            result = taxonomy_service.find_taxonomy_by_id(conn, entity_id)
        elif entity == "study":
            result = study_service.find_study_by_id(conn, entity_id)
        elif entity == "disease":
            result = disease_service.find_disease_by_id(conn, entity_id)
        elif entity == "chemicals_and_drugs":
            result = chemicals_and_drugs_service.find_by_id(conn, entity_id)
        elif entity == "diet_and_food":
            result = diet_and_food_service.find_by_id(conn, entity_id)
        elif entity =="association":
            result = association_service.find_by_id(conn, entity_id)
        elif entity =="mesh":
            result = mesh_service.find_by_id(conn,entity_id)
        else:
            raise ValueError("Unsupported entity type")
    result = dict(result)
    # result["entity_type"] = entity
    result = formt_items_json(result)
    return result

@entity_api.put("/update/{entity}/{entity_id}")
async def update_entity(entity: str, entity_id: str, updateData: dict):
    if "tags" in updateData and isinstance(updateData["tags"], list):
        updateData["tags"] = json.dumps(updateData["tags"], ensure_ascii=False)
    if "short_name" in updateData and isinstance(updateData["short_name"], list):
        updateData["short_name"] = json.dumps(updateData["short_name"], ensure_ascii=False)
    with get_engine().begin() as conn:
        if entity == "taxonomy":
            taxonomy_service.update_taxonomy(conn, entity_id, updateData)
        elif entity == "study":
            study_service.update_study(conn, entity_id, updateData) 
        elif entity == "disease":
            disease_service.update_disease(conn, entity_id, updateData)  
        elif entity == "chemicals_and_drugs":
            chemicals_and_drugs_service.update(conn, entity_id, updateData)
        elif entity == "diet_and_food":
            diet_and_food_service.update(conn, entity_id, updateData)
        elif entity == "association":
            association_service.update(conn, entity_id, updateData)
            find_data = association_service.find_by_id(conn, entity_id)
            find_association= dict(find_data)
            entity_,relationship = association_service.get_entity(conn,find_association)
            graph_service.create_node_relation(entity_,relationship)
        elif entity =="mesh":
            mesh_service.update(conn, entity_id, updateData)
        else:
            raise ValueError("Unsupported entity type")
    if entity != "association": 
        find_node = graph_service.find_entity( entity, entity_id=entity_id)  # 同步更新图数据库中的节点信息
        if find_node:
            graph_service.update_entity(find_node, updateData)
    return {"message": "Entity updated successfully"}

@entity_api.post("/add/{entity}")
async def add_entity(entity: str, data: dict):
    if "tags" in data and isinstance(data["tags"], list):
        data["tags"] = json.dumps(data["tags"], ensure_ascii=False)
    if "short_name" in data and isinstance(data["short_name"], list):
        data["short_name"] = json.dumps(data["short_name"], ensure_ascii=False)
    with get_engine().begin() as conn:
        if entity=="taxonomy":  
            taxonomy_service.add_taxonomy(conn, data)
        elif entity == "study":
            study_service.add_study(conn, data)     
        elif entity == "disease":   
            disease_service.add_disease(conn, data)
        elif entity == "chemicals_and_drugs":
            chemicals_and_drugs_service.add(conn, data)
        elif entity == "diet_and_food":
            diet_and_food_service.add(conn, data)
        elif entity == "association":
            entity_id = association_service.add(conn, data)
            find_data = association_service.find_by_id(conn, entity_id)
            find_association= dict(find_data)
            entity,relationship = association_service.get_entity(conn,find_association)
            graph_service.create_node_relation(entity,relationship)
            pass
        elif entity =="mesh":
            mesh_service.add_mesh_node(conn, AddMeshNode(**data))
        else:
            raise ValueError("Unsupported entity type")
    return {"message": "Entity added successfully"}



@entity_api.get("/find-by-name/{entity}/{keywords}")
async def get_entity(entity: str, keywords: str):
    with get_engine().begin() as conn:
        if entity=="taxonomy":  
            result = taxonomy_service.find_taxonomy_by_keywords(conn, keywords)
        elif entity == "study":
            result = study_service.find_study_by_keywords(conn, keywords)
        elif entity == "disease":
            result = disease_service.find_disease_by_keywords(conn, keywords)
        elif entity == "chemicals_and_drugs":
            result = chemicals_and_drugs_service.find_by_keywords(conn, keywords)
        elif entity == "diet_and_food":
            result = diet_and_food_service.find_by_keywords(conn, keywords)
        elif entity =="association":
            result = association_service.find_by_keywords(conn, keywords)
        elif entity =="mesh":
            result = mesh_service.find_by_keywords(conn,keywords)
        else:
            raise ValueError("Unsupported entity type")
    return result



@entity_api.post("/details/{entity}/{entity_id}")
async def get_entity(entity: str, entity_id: str,
                     detailsNodeQuery: DetailsNodeQuery):
    with get_engine().begin() as conn:
        if entity=="taxonomy":  
            result = taxonomy_service.details_taxonomy_by_id(conn, entity_id)
        elif entity == "study":
            result = study_service.details_study_by_id(conn, entity_id)
        elif entity == "disease":
            result = disease_service.details_disease_by_id(conn, entity_id)
        elif entity == "chemicals_and_drugs":
            result = chemicals_and_drugs_service.details_by_id(conn, entity_id)
        elif entity == "diet_and_food": 
            result = diet_and_food_service.details_by_id(conn, entity_id)
        elif entity =="association":
            result = graph_service.get_association_details(entity_id)
        elif entity =="mesh":
            result = mesh_service.details_by_id(conn,entity_id)
        else:
            raise ValueError("Unsupported entity type")
    if type(result) != dict:
        # raise HTTPException(status_code=404, detail=f"{entity} with id {entity_id} not found")
        result = dict(result)
    # result["entity_type"] = entity
    nodes = detailsNodeQuery.nodes
    if entity not in nodes:
        nodes.append(entity)
    if "study" not in nodes:
        nodes.append("study")
    graph_table = graph_service.find_associations_by_entity_id(entity_type=entity,entity_id=entity_id,nodes=detailsNodeQuery.nodes)
    if len(graph_table)==1:
        result["graph_table"] = graph_table[0]
    else:
        raise HTTPException(status_code=404, detail=f"Graph data for {entity} with id {entity_id} not found or many records found") 

    return result

@entity_api.post("/import/{entity}")
async def import_entitys(entity: str, entity_list: list[dict]):
    with get_engine().begin() as conn:
        if entity == "taxonomy":
            result = taxonomy_service.import_taxonomy(conn, entity_list)
        elif entity == "study":
            result = study_service.import_studies(conn, entity_list)
        elif entity == "disease":
            result = disease_service.import_diseases(conn, entity_list)
        elif entity == "chemicals_and_drugs":
            result = chemicals_and_drugs_service.imports(conn, entity_list)
        elif entity == "diet_and_food":
            result = diet_and_food_service.imports(conn, entity_list)
        elif entity == "association":
            result = association_service.imports(conn, entity_list)
        elif entity =="mesh":
            result = mesh_service.imports(conn,entity_list)
        else:
            raise ValueError("Unsupported entity type")
    return result


@entity_api.delete("/delete/{entity}/{entity_id}")
async def delete_entity(entity: str, entity_id: str,force: bool=False):
   
    find_entity = graph_service.find_entity( entity, entity_id)
    if find_entity:
        raise HTTPException(status_code=400, detail=f"Cannot delete {entity}({entity_id}) that exists in graph database")
        # return {"message": "Cannot delete entity that exists in graph database"}
    else: 
        with get_engine().begin() as conn:
            if entity == "taxonomy":
                taxonomy_service.delete_taxonomy_by_id(conn, entity_id)
            elif entity == "study":
                study_service.delete_study_by_id(conn, entity_id)
            elif entity == "disease":
                disease_service.delete_disease_by_id(conn, entity_id)
            elif entity == "chemicals_and_drugs":
                chemicals_and_drugs_service.delete_by_id(conn, entity_id)
            elif entity == "diet_and_food":
                diet_and_food_service.delete_by_id(conn, entity_id)
            # elif entity == "association":
            #     association_service.delete_by_id(conn, entity_id)
            elif entity =="mesh":
                mesh_service.delete_by_id(conn,entity_id)
            else:
                raise ValueError("Unsupported entity type")
        return {"message": "Entity deleted successfully"}

@entity_api.delete("/delete-node/{entity}/{entity_id}")
async def delete_entity_node(entity: str, entity_id: str):
    graph_service.delete_node_by_entity_id(entity, entity_id)
    return {"message": "Entity node and its relationships deleted successfully"}


@entity_api.get("/nodes")
async def get_all_nodes():
    graph = get_graph()
    
    # 查询所有节点
    query = """
    MATCH (n)
    RETURN id(n) AS id, labels(n) AS labels, n.entity_id AS entity_id, n.entity_name AS entity_name, properties(n) AS props
    """
    results = graph.run(query).data()
    
    return {
        "total": len(results),
        "nodes": results
    }

@entity_api.delete("/node-id/{node_id}")
def delete_node_by_id(node_id: int):
    graph_service.delete_by_node_id(node_id)
    return {"message": f"Node {node_id} and its relationships deleted successfully"}


@entity_api.post("/query-nodes")
def query_nodes(query: NodeQuery):
    return graph_service.query_nodes(query)




from operator import and_, or_
from shortuuid import uuid
from brave.microbe.schemas.entity import PageEntity
from brave.microbe.models.core import t_taxonomy
from sqlalchemy import select,func,insert,case,exists
from brave.microbe.utils.import_ncbi_taxonomy import merge_all

from sqlalchemy.engine import Connection

def page_taxonomy(conn: Connection, query: PageEntity):
    stmt =select(
        t_taxonomy,
    ) 
    
    # Left join with t_namespace to get namespace name
    # stmt = stmt.select_from(
    #    t_pipeline_components.outerjoin(t_namespace, t_pipeline_components.c.namespace == t_namespace.c.namespace_id)
    # )

    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
            # or_(
               
            # )
             t_taxonomy.c.scientific_name.ilike(keyword_pattern)
        )
    
    if query.parent_id:
        conditions.append(
            t_taxonomy.c.parent_tax_id == query.parent_id
        )
    

    # stmt = stmt.where(and_(*conditions))
    # count_stmt = select(func.count()).select_from(t_taxonomy).where(and_(*conditions))
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_taxonomy).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_taxonomy)

    stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_taxonomy = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_taxonomy,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def page_taxonomy_v2(conn: Connection, query: PageEntity):
    child = t_taxonomy.alias("child")

    stmt = (
        select(
            t_taxonomy,
            func.count(child.c.tax_id).label("child_count")
        )
        .select_from(
            t_taxonomy.outerjoin(child, t_taxonomy.c.tax_id == child.c.parent_tax_id)
        )
        .group_by(t_taxonomy.c.tax_id)
    )
    # Left join with t_namespace to get namespace name
    # stmt = stmt.select_from(
    #    t_pipeline_components.outerjoin(t_namespace, t_pipeline_components.c.namespace == t_namespace.c.namespace_id)
    # )

    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
            # or_(
               
            # )
             t_taxonomy.c.scientific_name.ilike(keyword_pattern)
        )
    
    if query.parent_id:
        conditions.append(
            t_taxonomy.c.parent_tax_id == query.parent_id
        )
    

    # stmt = stmt.where(and_(*conditions))
    # count_stmt = select(func.count()).select_from(t_taxonomy).where(and_(*conditions))
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_taxonomy).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_taxonomy)

    stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_taxonomy = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_taxonomy,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def page_taxonomy_v3(conn: Connection, query: PageEntity):
    child = t_taxonomy.alias("child")
    if query.parent_id and query.parent_id=="default":
        query.parent_id = 1
    stmt = (
        select(
            t_taxonomy,
            case(
                (exists().where(child.c.parent_tax_id == t_taxonomy.c.tax_id), True),
                else_=False
            ).label("has_children")
        )
    )
    # Left join with t_namespace to get namespace name
    # stmt = stmt.select_from(
    #    t_pipeline_components.outerjoin(t_namespace, t_pipeline_components.c.namespace == t_namespace.c.namespace_id)
    # )

    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
            # or_(
               
            # )
             t_taxonomy.c.scientific_name.ilike(keyword_pattern)
        )
    
    if query.parent_id:
        conditions.append(
            t_taxonomy.c.parent_tax_id == query.parent_id
        )
    

    # stmt = stmt.where(and_(*conditions))
    # count_stmt = select(func.count()).select_from(t_taxonomy).where(and_(*conditions))
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_taxonomy).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_taxonomy)
    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_taxonomy = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    return {
        "items": find_taxonomy,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }



def find_taxonomy_by_id(conn: Connection, entity_id: str):
    stmt = t_taxonomy.select().where(t_taxonomy.c.entity_id ==entity_id)
    find_taxonomy = conn.execute(stmt).mappings().first()
    return find_taxonomy

def update_taxonomy(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_taxonomy.update().where(t_taxonomy.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)

def add_taxonomy(conn: Connection, data: dict):
    data['entity_id'] = str(uuid())
    stmt = t_taxonomy.insert().values(data)
    conn.execute(stmt)

def find_taxonomy_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_taxonomy.select().where(t_taxonomy.c.entity_name.ilike(keyword_pattern)).limit(10)
    find_taxonomy = conn.execute(stmt).mappings().all()
    return find_taxonomy


def details_taxonomy_by_id(conn: Connection, entity_id: str):
    stmt = t_taxonomy.select().where(t_taxonomy.c.entity_id ==entity_id)
    find_taxonomy = conn.execute(stmt).mappings().first()
    lineage = get_lineage(find_taxonomy.tax_id, conn)
    find_taxonomy = dict(find_taxonomy)
    find_taxonomy['lineage'] = build_lineage_chain(lineage, find_taxonomy['tax_id'])
    
    return find_taxonomy


def build_lineage_chain(result, target_tax_id):
    # 1. 建立 tax_id -> node 映射
    tax_map = {row.tax_id: row for row in result}

    # 2. 从目标 tax_id 开始向上查
    lineage = []
    current_id = target_tax_id

    while current_id in tax_map:
        node = tax_map[current_id]
        lineage.append({
            "tax_id": node.tax_id,
            "parent_tax_id": node.parent_tax_id,
            "entity_name": node.entity_name,
            "rank": node.rank
        })
        # parent_tax_id = 1 停止
        if node.parent_tax_id == 1:
            break
        current_id = node.parent_tax_id

    # 3. 反转顺序：root -> 当前节点
    lineage.reverse()
    return lineage

def get_lineage(tax_id: int, conn):
    taxonomy = t_taxonomy.alias()

    # 递归 CTE
    lineage = (
        select(
            taxonomy.c.tax_id,
            taxonomy.c.parent_tax_id,
            taxonomy.c.entity_name,
            taxonomy.c.rank
        )
        .where(taxonomy.c.tax_id == tax_id)
        .cte(name="lineage", recursive=True)
    )

    parent = taxonomy.alias("parent")
    lineage = lineage.union_all(
        select(
            parent.c.tax_id,
            parent.c.parent_tax_id,
            parent.c.entity_name,
            parent.c.rank
        ).where(
            (parent.c.tax_id == lineage.c.parent_tax_id) &
            (lineage.c.parent_tax_id != 1)  # parent_tax_id = 1 停止递归
        )
    )

    stmt = select(lineage).order_by(lineage.c.tax_id)
    result = conn.execute(stmt).fetchall()
    return result


def import_taxonomy(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    for i in range(0, len(records), batch_size):
        batch = records[i:i + batch_size]
        stmt = insert(t_taxonomy)
        conn.execute(stmt, batch)
        inserted += len(batch)
        print(f"Inserted {inserted} records...")
    return inserted

def delete_taxonomy_by_id(conn: Connection, entity_id: str):
    stmt = t_taxonomy.delete().where(t_taxonomy.c.entity_id == entity_id)
    conn.execute(stmt)


def  init(conn: Connection):
    with open("brave/microbe/data/taxonomy.sql", "r") as f:
        taxonomy = merge_all(
            "/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/nodes.dmp",
            "/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/names.dmp",
            "/ssd1/wy/workspace2/nextflow-fastapi/taxonomy/division.dmp"
        )
        # taxonomy["taxonomy_id"] = f"TAX{taxonomy['tax_id']}"
        taxonomy["taxonomy_id"] = "TAX" + taxonomy["tax_id"].astype(str)
        records = taxonomy.to_dict(orient="records")
        batch_size = 20000  # 每次 1w 行，自己调节

        inserted = 0
        for i in range(0, len(records), batch_size):
            batch = records[i:i + batch_size]
            stmt = insert(t_taxonomy)
            conn.execute(stmt, batch)
            inserted += len(batch)
            print(f"Inserted {inserted} records...")

    return {"message": f"导入完成，共导入 {inserted} 行"}



from typing import Callable, Awaitable, Dict, Union, Coroutine
import asyncio
from collections import defaultdict

from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.event import AnalysisResultEvent
from brave.api.schemas.analysis import Analysis, AnalysisExecuterModal
from brave.api.schemas.analysis_result import AnalysisResult, AnalysisResultParseModal

Callback = Union[Callable[[Analysis,AnalysisResultParseModal], None], Callable[[Analysis,AnalysisResultParseModal], Coroutine]]

class AnalysisResultRouter(BaseEventRouter[AnalysisResultEvent,Callback]):
    async def dispatch(self, event: AnalysisResultEvent, analsyis: Analysis,analysis_result:AnalysisResultParseModal):
        handlers = self._handlers.get(event, [])
 
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler(analsyis,analysis_result )
            else:
                handler(analsyis,analysis_result)

    


from anyio import Event
from dependency_injector.wiring import Provide
from dependency_injector.wiring import inject
from fastapi import APIRouter,Depends,HTTPException, Request
from sqlalchemy.orm import Session
# from brave.api.config.db import conn
from brave.api.core.evenet_bus import EventBus
from brave.api.executor.models import LocalJobSpec
from brave.api.schemas.bio_database import QueryBiodatabase
from brave.api.schemas.sample import Sample
from typing import List
from starlette.status import HTTP_204_NO_CONTENT
from sqlalchemy import func, select
from brave.api.models.orm import SampleAnalysisResult
import glob
import importlib
import os
from brave.api.config.db import get_db_session
from sqlalchemy import and_,or_
import pandas as pd
from brave.api.models.core import samples
from brave.api.config.db import get_engine
from brave.api.schemas.analysis import AnalysisInput,Analysis, UpdateProject,QueryAnalysis,AnalysisExecuterModal
from typing import Dict, Any
from brave.api.models.core import analysis,t_container
import json
import importlib
import importlib.util
import uuid
import os
from brave.api.service.result_parse import script_analysis
from brave.api.service.result_parse import nextflow_analysis
from brave.api.service.result_parse.nextflow_analysis    import NextflowAnalysis
from brave.api.service.result_parse.result_parse import ResultParse
from brave.api.utils.get_db_utils import get_ids
from brave.api.config.config import get_settings
from brave.api.routes.pipeline import get_pipeline_file
import textwrap
# from brave.api.routes.sample_result import find_analyais_result_by_ids
from brave.api.routes.sample_result import parse_result_one
import  brave.api.service.pipeline as pipeline_service
import brave.api.service.bio_database_service as bio_database_service
from typing import Optional
import pandas as pd
import subprocess
from brave.api.service.watch_service import queue_process
import threading
import psutil
import brave.api.service.analysis_result_service as analysis_result_service
import brave.api.service.sample_service as sample_service
import brave.api.service.analysis_service as analysis_service
from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.app_container import AppContainer
from brave.app_manager import AppManager
from brave.api.executor.base import JobExecutor
from brave.api.core.routers_name import RoutersName
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.service.result_parse.script_analysis import ScriptAnalysis
from brave.api.service.result_parse.nextflow_analysis import NextflowAnalysis
from brave.api.utils.file_utils import delete_all_in_dir
import  brave.api.service.file_operation  as file_operation_service
import pandas as pd
import brave.api.service.container_service as container_service
import shutil
import brave.api.service.notebook as notebook_service
from brave.api.models.core import t_pipeline_components
from sqlalchemy.orm import aliased
from collections import defaultdict

analysis_api = APIRouter()



def update_or_save_result(db,project,sample_name,file_type,file_path,log_path,verison,analysis_name,software):
        sampleAnalysisResult = db.query(SampleAnalysisResult) \
        .filter(and_(SampleAnalysisResult.analysis_name == analysis_name,\
                SampleAnalysisResult.analysis_verison == verison, \
                SampleAnalysisResult.sample_name == sample_name, \
                SampleAnalysisResult.file_type == file_type, \
                SampleAnalysisResult.project == project \
            )).first()
        if sampleAnalysisResult:
            sampleAnalysisResult.contant_path = file_path
            sampleAnalysisResult.log_path = log_path
            sampleAnalysisResult.software = software
            db.commit()
            db.refresh(sampleAnalysisResult)
            print(">>>>更新: ",file_path,sample_name,file_type,log_path)
        else:
            sampleAnalysisResult = SampleAnalysisResult(analysis_name=analysis_name, \
                analysis_verison=verison, \
                sample_name=sample_name, \
                file_type=file_type, \
                log_path=log_path, \
                software=software, \
                project=project, \
                contant_path=file_path \
                    )
            db.add(sampleAnalysisResult)
            db.commit()
            print(">>>>新增: ",file_path,sample_name,file_type,log_path)


# def get_db_value(session, value):
#     ids = value
#     if not isinstance(value,list):
#         ids = [value]
#     analysis_result =  session.query(SampleAnalysisResult) \
#                 .filter(SampleAnalysisResult.id.in_(ids)) \
#                     .all()
                    
#     for item in analysis_result:
#         if item.content_type=="json" and not isinstance(item.content, dict):
#             item.content = json.loads(item.content)

#     if len(analysis_result)!=len(ids):
#         raise HTTPException(status_code=500, detail="数据存在问题!")
#     if not isinstance(value,list) and len(analysis_result)==1:
#         return analysis_result[0]
#     else:
#         return analysis_result
    




 

        # get_script = getattr(module, "get_script")
        # script = get_script()
        # command = f"nextflow run -offline {script} -resume  -params-file {params_path} -w {work_dir} -with-trace trace.txt | tee .workflow.log"
        # with open(command_path, "w") as f:
        #     f.write(command)
        # get_output_format = getattr(module, "get_output_format")
        # output_format = get_output_format()
        # return json.dumps(output_format)

# ,response_model=List[Sample]
#  参数解析


def get_all_files_recursive_v2(directory):
    file_list=[]
    for root, dirs, files in os.walk(directory):
        for file in files:
            file_list.append(os.path.join(root, file).replace(directory,""))
    return file_list




@analysis_api.get("/analysis/browse-output-dir/{analysis_id}")
async def browse_output_dir(analysis_id):
   with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt).mappings().first()
        if not result:
            raise HTTPException(status_code=404, detail=f"Analysis with id {analysis_id} not found")
        output_dir = result['output_dir']
        file_list = get_all_files_recursive_v2(output_dir)
        return file_list

# 结果解析

@analysis_api.post("/fast-api/parse-analysis-result/{analysis_id}")
@inject
async def parse_analysis_result(
    analysis_id,save:Optional[bool]=False,
    event_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
   ):


    resultParse = ResultParse(analysis_id,event_bus)
    
    with get_engine().begin() as conn:
        if save:
            result_list,result_dict,params = await resultParse.parse(True)
            # result = await analysis_result_parse_service.save_analysis_result_preview(conn,analysis_id)
        else:
            # result = await analysis_result_parse_service.parse_analysis_result_preview(conn,analysis_id)
            result_list,result_dict,params = await resultParse.parse(False)

        file_format_list = params["file_format_list"]
        file_dict = analysis_service.get_file_dict(file_format_list,params['analysis']['output_dir'])
        # file_dict = find_file_dict(file_format_list,params['analysis']['output_dir'])   
        return {"result_dict":result_dict,"file_format_list":file_format_list,"file_dict":file_dict}


@analysis_api.post(
    "/list-analysis",
    # response_model=List[Analysis],
)
async def list_analysis(query:QueryAnalysis):
    with get_engine().begin() as conn:
        return  analysis_service.list_analysis(conn,query)

def build_tree(data):
    grouped = defaultdict(list)
    for item in data:
        grouped[item["component_id"]].append(item)

    tree = []
    for comp_id, items in grouped.items():
        parent = {
            "title": items[0]["component_name"],  # 每组的名字
            "key": comp_id,   
            "type":"components",                    # 用 component_id 当 key
            "children": []
        }
        for i, it in enumerate(items):
            parent["children"].append({
                "title": it["analysis_name"],
                "type":"analysis",
                "key": it["analysis_id"],
                "component_id": it["component_id"],
                "component_name": it["component_name"],
                "component_type": it["component_type"],
                "analysis_status": it["analysis_status"],
                "component_order_index": it["component_order_index"]

            })
        tree.append(parent)
    return tree

@analysis_api.post(
    "/list-analysis-tree",
    # response_model=List[Analysis],
)
async def list_analysis(query:QueryAnalysis):
    with get_engine().begin() as conn:
        analysis_list = analysis_service.list_analysis(conn,query)
    analysis_list = [dict(item) for item in analysis_list]
    sorted_data = sorted(
        analysis_list,
        key=lambda x: (
            0 if x["component_order_index"] not in (None, "") else 1,   # 有值优先，空值最后
            -(int(x["component_order_index"])) if x["component_order_index"] not in (None, "") else float("inf")  # 数字取负 → 实现降序
        )
    )
    data = build_tree(sorted_data)
    return data


@analysis_api.delete("/fast-api/analysis/{analysis_id}",  status_code=HTTP_204_NO_CONTENT)
def delete_analysis(analysis_id: str):
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
        find_analysis_result = analysis_result_service.find_analysis_result_by_analysis_id(conn,analysis_id)
        if find_analysis_result:
            raise HTTPException(status_code=500, detail=f"存在分析结果不能删除!")
        output_dir = find_analysis.output_dir
        # delete_all_in_dir(output_dir)
        shutil.rmtree(output_dir)
        print(f"删除文件夹: {output_dir}")
        conn.execute(analysis.delete().where(analysis.c.analysis_id == analysis_id))

    return {"message":"success"}




def pileine_analysis_run_log(result,type):
    if type == "workflow_log":
        workflow_log_file = result.workflow_log_file
        if workflow_log_file and os.path.exists(workflow_log_file):
            with open(workflow_log_file, "r") as f:
                params =f.read()
            return params
    elif type == "executor_log":
        executor_log_file = result.executor_log_file
        if executor_log_file and os.path.exists(executor_log_file):
            with open(executor_log_file, "r") as f:
                params =f.read()
            return params
    elif type == "params":
        params_path = result.params_path
        if params_path and os.path.exists(params_path):
            with open(params_path, "r") as f:
                params = json.load(f)
            return params
    elif type == "script_config":
        script_config_file = result.script_config_file
        if script_config_file and os.path.exists(script_config_file):
            with open(script_config_file, "r") as f:
                params = f.read()
            return params
    elif type == "trace":
        trace_file = result.trace_file
        trace =[]
        total = 0
        if trace_file and os.path.exists(trace_file):
            df = pd.read_csv(trace_file,sep="\t")
            total = df.shape[0]
            trace = df.to_dict(orient="records")
        
        return {
            "traceTable":trace,
            "total":total,
            "process_id":result.process_id,
            "status":"running" if result.process_id else "finished"
        }
    return ""


@analysis_api.get("/monitor-analysis/{analysis_id}")
async def pipeline_monitor(analysis_id,type):
    with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt)
        result = result.mappings().first()
    return pileine_analysis_run_log(result,type)
    # analysis_ = rows[len(rows)-1]
    # output_dir = analysis_['output_dir']
    
# import asyncio
# import time
# import threading
# def blocking_task():
#     print(f"开始阻塞任务，线程: {threading.current_thread().name}")
#     time.sleep(5)
#     print("阻塞任务完成")

def start_background( cwd,cmd):
    proc = subprocess.Popen(
        cmd,
        cwd=cwd, 
        start_new_session=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )
    threading.Thread(target=proc.wait, daemon=True).start() # 处理僵尸进程
    return proc.pid

@analysis_api.post("/run-analysis/{analysis_id}")
@inject
async def run_analysis(
    request: Request,
    analysis_id,
    auto_parse:Optional[bool]=True,
    analysis_result_parse_service:AnalysisResultParse = Depends(Provide[AppContainer.analysis_result_parse_service])         
    ):

    manager: AppManager = request.app.state.manager  # 从 app.state 获取实例
    process_monitor = manager.process_monitor
    if process_monitor is None:
        raise HTTPException(status_code=500, detail="ProcessMonitor服务未初始化")
    
    with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt)
        analysis_ = result.mappings().first()
        process_id = analysis_.process_id
        if process_id is not None:
            try:
                proc = psutil.Process(int(process_id))
                if proc.is_running():
                    raise Exception(f"Analysis is already running with process_id={process_id}")
            except (psutil.NoSuchProcess, ValueError):
                pass  # 进程不存在或 process_id 非法，继续执行
        
        pid = start_background(analysis_.output_dir, ["bash","run.sh"])
        stmt = analysis.update().values({"process_id":pid,"analysis_status":"running"}).where(analysis.c.analysis_id==analysis_id)
        conn.execute(stmt)
        analysis_dict = dict(analysis_)

        analysis_dict['process_id'] = pid
        # await queue_process.put(analysis_dict)
        await process_monitor.add_process(analysis_dict)
        if auto_parse:
            await analysis_result_parse_service.add_analysis_id(analysis_id)
    return {"pid":pid}




# @analysis_api.post("/fast-api/save-analysis")
# @inject
# async def save_analysis(
#     request_param: Dict[str, Any],
#     save:Optional[bool]=False,
#     is_submit:Optional[bool]=False,
#     software_analysis:NextflowAnalysis  =Depends(Provide[AppContainer.nextflow_analysis])): # request_param: Dict[str, Any]
#     with get_engine().begin() as conn:
#         parse_analysis_result,component = software_analysis.get_parames(conn,request_param)
#         if not save:
#             return parse_analysis_result
#         return await software_analysis.save_analysis(conn,request_param,parse_analysis_result,component,is_submit) 
  
    # return software_analysis.save_analysis(request_param)
    # return {"msg":"success"}


@analysis_api.post("/fast-api/analysis-controller")
@inject
async def save_script_analysis(
    request_param: Dict[str, Any],
    # type:Optional[str]="nextflow",
    save:Optional[bool]=False,
    is_submit:Optional[bool]=False,
    app_container:AppContainer = Depends(Provide[AppContainer])
    ): # request_param: Dict[str, Any]
    

    with get_engine().begin() as conn:
        component_id = request_param['component_id']
        # pipeline_id = request_param['pipeline_id']
        if component_id is None:
            raise HTTPException(status_code=500, detail=f"component_id is None")
        component = pipeline_service.find_pipeline_by_id(conn, component_id)
        if component["component_type"] == "pipeline":
            component = pipeline_service.get_pipeline_v2(conn,component_id)
        if component is None:
            raise HTTPException(status_code=404, detail=f"Component with id {component_id} not found")
        if component is None or "content" not in component:
            raise HTTPException(status_code=404, detail=f"Component with id {component_id} not found or missing content.")
        
        # component_content = 
        component_obj = {
            **{ k:v for k,v in component.items() if k != "content"},
            **json.loads(component['content'])
        }

        script_type = component_obj['script_type']   
        # if  script_type=="nextflow":
        #     analysis_controller = app_container.nextflow_analysis()
        # else:
        #     analysis_controller = app_container.script_analysis()
        

        if script_type == "python" or script_type == "shell" or script_type == "r" or script_type == "jupyter":
            analysis_controller = app_container.script_analysis()
        else:
            analysis_controller = app_container.nextflow_analysis()

        parse_analysis_result = analysis_controller.get_parames(conn,request_param,component_obj)

        if not save:
            return parse_analysis_result
        return await analysis_controller.save_analysis(conn,request_param,parse_analysis_result,component_obj,is_submit) 
  


@analysis_api.get("/get-executor-logs/{analysis_id}")
@inject
def get_executor_logs(analysis_id,job_executor_selector:JobExecutor = Depends(Provide[AppContainer.job_executor_selector])):
    return job_executor_selector.get_logs(analysis_id)




@analysis_api.post("/run-analysis-v2/{analysis_id}")
@inject
async def run_analysis_v2(
    analysis_id,
    run_type:str="job",
    auto_parse:Optional[bool]=True,
    # executor: JobExecutor = Depends(get_executor_dep),
    evenet_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
    ):

    # manager: AppManager = request.app.state.manager  # 从 app.state 获取实例
    # process_monitor = manager.process_monitor
    # if process_monitor is None:
    #     raise HTTPException(status_code=500, detail="ProcessMonitor服务未初始化")
    
    with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt)
        analysis_ = result.mappings().first()
        if analysis_ is None:
            raise HTTPException(status_code=404, detail="Analysis not found")
        # process_id = analysis_['process_id']
        component = pipeline_service.find_component_by_id(conn,analysis_["component_id"])
        component_type = component['component_type']
        if component_type=="script":
            output_dir = f"{analysis_['output_dir']}/output"
            # if os.path.exists(output_dir):
            delete_all_in_dir(output_dir)
        
        if not component["container_id"]:
            raise HTTPException(status_code=500, detail=f"please config container id") 

        # find_container = container_service.find_container_by_id(conn,analysis_["container_id"])
        analysis_ = dict(analysis_)
        analysis_["run_id"] = f"{run_type}-{analysis_id}"
        
        if run_type == "job":
            analysis_["container_id"] =component["container_id"]
        else:
            # if component_type=="script":
            analysis_["container_id"] =component["container_id"]
            # else: 
            #     if not component["sub_container_id"]:
            #         raise HTTPException(status_code=500, detail=f"please config sub_container_id id") 
            #     analysis_["container_id"] = component["sub_container_id"]

        # analysis_["image"] = find_container["image"]
        analysis_ = AnalysisExecuterModal(**analysis_)
        # analysis_.image = find_container["image"]
        await analysis_service.finished_analysis(analysis_id,run_type,"running")
        # stmt = analysis.update().values({"analysis_status":"running","run_type":run_type}).where(analysis.c.analysis_id==analysis_id)
        conn.execute(stmt)
    await evenet_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_SUBMITTED,analysis_)
    
    
    # job_id = await executor.submit_job(LocalJobSpec(
    #     job_id=analysis_id,
    #     command=["bash", "run.sh"],
    #     output_dir=analysis_['output_dir'],
    #     process_id=analysis_['process_id']
    # ))

    return {"msg":"success"}


@analysis_api.post("/analysis/stop-analysis/{analysis_id}")
@inject
async def stop_analysis(
    analysis_id,
    run_type,
    evenet_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
    ):


    with get_engine().begin() as conn:
        stmt = select(analysis).where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt)
        analysis_ = result.mappings().first()
        if analysis_ is None:
            raise HTTPException(status_code=404, detail="Analysis not found")
        
        run_id = f"{run_type}-{analysis_id}"
        analysis_ = AnalysisExecuterModal(**analysis_,run_id=run_id)
        
        # stmt = analysis.update().values({"analysis_status":"running"}).where(analysis.c.analysis_id==analysis_id)
        # conn.execute(stmt)

        await evenet_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_STOPED,analysis_)
        
    
        return {"msg":"success"}


@analysis_api.get("/find-analysis-by-id/{analysis_id}") 
async def find_analysis_by_id(analysis_id):
    settings = get_settings()
    with get_engine().begin() as conn:
        t_sub_container = aliased(t_container)

        stmt = select(
            analysis,
            t_container.c.name.label("container_name"),
            t_container.c.image.label("container_image"),
            t_sub_container.c.name.label("sub_container_name"),
            t_sub_container.c.image.label("sub_container_image")
        )

        stmt = stmt.select_from(
                analysis.outerjoin(t_pipeline_components,analysis.c.component_id==t_pipeline_components.c.component_id)
                .outerjoin(t_container,t_pipeline_components.c.container_id==t_container.c.container_id)
                .outerjoin(t_sub_container,t_pipeline_components.c.sub_container_id==t_sub_container.c.container_id)
            #  analysis.outerjoin(t_container, analysis.c.container_id == t_container.c.container_id)
        )
        stmt = stmt.where(analysis.c.analysis_id == analysis_id)
        result = conn.execute(stmt)
        analysis_ = result.mappings().first()
        if analysis_ is None:
            raise HTTPException(status_code=404, detail="Analysis not found")
        analysis_dict = dict(analysis_)
        script_dir = analysis_dict['pipeline_script']
        script_dir = os.path.dirname(script_dir).replace(str(settings.PIPELINE_DIR)+"/","")
        jupyter_notebook_path =  f"{script_dir}/main.ipynb"
        analysis_dict["jupyter_notebook_path"] = jupyter_notebook_path
        
        return analysis_dict



@analysis_api.get("/get-cache-analysis-result-by-id/{analysis_id}")
@inject
async def get_cache_analysis_result_by_id(analysis_id,analysis_result_parse_service:AnalysisResultParse = Depends(Provide[AppContainer.analysis_result_parse_service])):
    return analysis_result_parse_service.cached_analysis_result()[analysis_id]

@analysis_api.get("/get-cache-analysis-result")
@inject
async def get_cache_analysis_result(analysis_result_parse_service:AnalysisResultParse = Depends(Provide[AppContainer.analysis_result_parse_service])):
    return analysis_result_parse_service.cached_analysis_result()


@analysis_api.get("/get-cache-analysis-params")
@inject
async def get_cache_params(analysis_result_parse_service:AnalysisResultParse = Depends(Provide[AppContainer.analysis_result_parse_service])):
    return analysis_result_parse_service.cached_params()



@analysis_api.get("/analysis/visualization-results/{analysis_id}")
async def visualization_results(analysis_id):
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
        find_component = pipeline_service.find_component_by_id(conn,find_analysis['component_id'])
    file_result = await file_operation_service.visualization_results(find_analysis["output_dir"])
    # file_result = {}
    file_result['description'] = find_component["description"]
    file_result['analysis_name'] = find_analysis["analysis_name"]
    file_result['job_status'] = find_analysis["job_status"]
    file_result['server_status'] = find_analysis["server_status"]
    file_result['is_report'] = find_analysis["is_report"]
    file_result['analysis_id'] = find_analysis["analysis_id"]

    file_result['component_name'] = find_component["component_name"]
    file_result['component_id'] = find_component["component_id"]

    file_result['component_type'] = find_component["component_type"]
    file_result['command_log_path'] = find_analysis["command_log_path"]


    return file_result

@analysis_api.get("/analysis/analysis-progress/{analysis_id}")
async def analysis_progress(analysis_id):
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
    trace_file = find_analysis["trace_file"]
    df = pd.read_csv(trace_file,sep="\t")
    return json.loads(df.to_json(orient="records"))


@analysis_api.post("/analysis/convert-ipynb/{analysis_id}")
async def convert_ipynb(analysis_id):
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
    script_path  = find_analysis["pipeline_script"]
    ipynb_path = os.path.dirname(script_path)
    ipynb_path = f"{ipynb_path}/main.ipynb"
    if os.path.exists(script_path) and os.path.exists(ipynb_path):
        shutil.copy(script_path, f"{script_path}.tmp")
        notebook_service.convert_notebook(ipynb_path, script_path)
        return "success"
    raise HTTPException(status_code=404, detail=f"{script_path}或{ipynb_path}不存在!")


@analysis_api.post("/analysis/update-report/{analysis_id}")
async def update_report(analysis_id):
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
        if find_analysis['is_report']:
            analysis_service.update_report(conn,analysis_id,False)
        else:
            analysis_service.update_report(conn,analysis_id,True)
    return "success"


@analysis_api.post("/analysis/edit-params/{analysis_id}")
async def edit_params(analysis_id):
    analysis_result = {}
    inputFormJson = []
    # project = editParams.project if editParams.project else []
    with get_engine().begin() as conn:
        find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
        find_component = pipeline_service.find_component_by_id(conn,find_analysis['component_id'])
        project = json.loads(find_analysis["extra_project_ids"]) if "extra_project_ids" in  find_analysis and find_analysis["extra_project_ids"] else []

        request_param = json.loads(find_analysis["request_param"])
        if "data_component_ids" in request_param:
            data_component_ids = request_param["data_component_ids"]
            data_component_ids = json.loads(data_component_ids)
            # find_analysis["project"]
           
            project.append(find_analysis["project"])
            analysis_result = analysis_result_service.find_analyais_result_groupd_by_component_ids(conn,data_component_ids,project)
        
        component_type = find_component["component_type"]
        
        if component_type == "software" or component_type=="pipeline":
            software_input_file = pipeline_service.find_component_by_parent_id(conn,find_component['component_id'],"software_input_file")
            inputFormJson = [
                {
                    # **{k:v for k,v in item.items() if k!="content"},
                    "component_id":item["component_id"],
                    **json.loads(item["content"])
                } for item in software_input_file
            ]

    result = {
        "analysis_name":find_analysis["analysis_name"],
        "is_report":find_analysis["is_report"],
        "analysis_id":find_analysis["analysis_id"],
        "component_name":find_component["component_name"],
        "component_type":find_component["component_type"],
        "component_id":find_component["component_id"],
        "request_param":json.loads(find_analysis["request_param"]),
        "content":json.loads(find_component["content"]),
        "analysis_result":analysis_result,
        "inputFormJson":inputFormJson

    }

    return result


@analysis_api.post("/analysis/update-extra-project/{analysis_id}")
async def edit_params(project:UpdateProject,analysis_id):
    with get_engine().begin() as conn:
        # find_analysis = analysis_service.find_analysis_by_id(conn,analysis_id)
        analysis_service.update_extra_project(conn,analysis_id,project.project)
    return "success"


from collections import defaultdict
from enum import Enum
from typing import Callable, Awaitable, Dict, Union, Coroutine
import asyncio

from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.direct_dispatch import DirectDispatch
from brave.api.core.event import IngressEvent
from brave.api.core.queued_dispatch import QueuedDispatch


Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]

class IngressEventRouter(BaseEventRouter[IngressEvent,Callback]):
    async def dispatch(self, event: IngressEvent, payload: dict):
        handlers = self._handlers.get(event, [])
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler(payload)
            else:
                handler(payload)
    # def __init__(self):
    #     super().__init__(maxlen=100, delay=0.02)

    # def __init__(self):
    #     self._handlers: dict[IngressEvent, set[Callback]] = defaultdict(set)

    # def register_handler(self, event: IngressEvent, handler: Callback):
    #     self._handlers[event].add(handler)
    
    # async def dispatch(self, event: IngressEvent, msg: dict):
    #     # event_str = msg.get("ingress_event")
    #     # if not event_str:
    #     #     print("[IngressEventRouter] No 'ingress_event' in message")
    #     #     return
        
    #     # data = msg.get("data")
    #     # if not data:
    #     #     print("[IngressEventRouter] No 'data' in message")
    #     #     return
        
    #     # try:
    #     #     event = IngressEvent(event_str)
    #     # except ValueError:
    #     #     print(f"[IngressEventRouter] Unknown event type '{event_str}'", msg)
    #     #     return
        
    #     handlers = self._handlers.get(event)
    #     if handlers:
    #         for handler in handlers:
    #             if asyncio.iscoroutinefunction(handler):
    #                 await handler(msg)
    #             else:
    #                 handler(msg)
    #     else:
    #         print(f"[IngressEventRouter] No handler for event '{event}'")

# from .workflow_queue import WorkflowQueueManager

# class IngressEventRouter:
#     def __init__(self, wq_manager: WorkflowQueueManager):
#         self.wq_manager = wq_manager

#     async def dispatch(self, msg: dict):
#         workflow_id = msg.get("workflow_id", "global")
#         await self.wq_manager.put(workflow_id, msg)
    



from typing import Optional
from fastapi import APIRouter, Depends, HTTPException
from brave.api.config.config import get_settings
from brave.api.core.evenet_bus import EventBus
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.core.routers_name import RoutersName
from brave.api.executor.base import JobExecutor
from brave.api.schemas.analysis import AnalysisExecuterModal
from brave.api.schemas.container import ListContainerQuery, PageContainerQuery,SaveContainer
import brave.api.service.container_service as container_service
from brave.api.config.db import get_engine
from brave.api.config.db import get_engine
import brave.api.service.pipeline as pipeline_service
import uuid
from dependency_injector.wiring import Provide
from dependency_injector.wiring import inject
import asyncio

from brave.app_container import AppContainer

container_controller = APIRouter(prefix="/container")


async def format_container(job_executor, container_list):
    containers = await job_executor.list_running()
    containers_dict = {item["name"]: item["id"] for item in containers }
    items = []

    for item in container_list:
        container_id = item["container_id"]
        run_id = f"retry-{container_id}"
        if run_id in containers_dict:
            item["status"] = "running"
            item["docker_id"] = containers_dict[run_id]
        else:
            item["status"] = "stopped"
        items.append(item)
    return items

@container_controller.post("/page", tags=['container'])
@inject
async def page_container(query:PageContainerQuery,
                         job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    with get_engine().begin() as conn:
        page_list =  container_service.page_container(conn,query)

    page_list['items'] = await format_container(job_executor, page_list["items"])
    return page_list

@container_controller.get("/find-by-id/{container_id}", tags=['container'])
async def find_by_id(container_id):
    with get_engine().begin() as conn:
        return container_service.find_container_by_id(conn,container_id)




@container_controller.post("/add-or-update-container",tags=['container'])
@inject
async def save_container_controller(saveContainer:SaveContainer, job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    image = ""
    save_container_dict  = saveContainer.model_dump()
    with get_engine().begin() as conn:
        if saveContainer.container_id:
            find_container= container_service.find_container_by_id(conn,saveContainer.container_id)
            container_id = find_container.container_id 
            image = saveContainer.image or find_container.image
            image = job_executor.get_image(image)
            if image:
                save_container_dict.update({"image_id":image.id,"image_status":"exist"})
            else:
                save_container_dict.update({"image_status":"not_exist"})
            container_service.update_container(conn,saveContainer.container_id,save_container_dict)
        else:
            str_uuid = str(uuid.uuid4())
            save_container_dict["container_id"] = str_uuid
            container_id = str_uuid
            image = saveContainer.image
            image = job_executor.get_image(image)
            if image:
                save_container_dict.update({"image_id":image.id,"image_status":"exist"})
            else:
                save_container_dict.update({"image_status":"not_exist"})
            container_service.save_container(conn,save_container_dict)
  
    return {"message":"success"}

@container_controller.delete("/delete-container-by-id/{container_id}",tags=['container'])
async def delete_by_container_id(container_id:str):
    with get_engine().begin() as conn:
        find_container = container_service.find_container_by_id(conn,container_id)
        if find_container:
            find_component =  pipeline_service.find_component_by_container_id(conn,container_id)
            if find_component:
                raise HTTPException(status_code=400, detail=f"container {container_id} 存在组件，不能删除")
        container_service.delete_container(conn,container_id)
   
    return {"message":"success"}




@container_controller.post("/run-container/{container_id}")
@inject
async def run_container(
    container_id,
    # executor: JobExecutor = Depends(get_executor_dep),
    evenet_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
    ):
    settings = get_settings()

    base_dir = str(settings.BASE_DIR)
    run_id = f"retry-{container_id}"
    analysis_ =  AnalysisExecuterModal(
        analysis_id=container_id,
        container_id=container_id,
        output_dir=base_dir,
        pipeline_script=f"{base_dir}/run.sh",
        run_id=run_id
    )
    # try:
    # raise RuntimeError("oops!")
    await evenet_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_SUBMITTED,analysis_)
    # except Exception as e:
    #     raise HTTPException(status_code=400, detail=str(e))
    
    # job_id = await executor.submit_job(LocalJobSpec(
    #     job_id=analysis_id,
    #     command=["bash", "run.sh"],
    #     output_dir=analysis_['output_dir'],
    #     process_id=analysis_['process_id']
    # ))

    return {"msg":"success"}


@container_controller.post("/stop-container/{container_id}")
@inject
async def stop_container(
    container_id,
    # executor: JobExecutor = Depends(get_executor_dep),
    evenet_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
    ):
    settings = get_settings()

    data_dir = str(settings.DATA_DIR)
    run_id = f"retry-{container_id}"
    analysis_ =  AnalysisExecuterModal(
        analysis_id=container_id,
        container_id=container_id,
        output_dir=data_dir,
        pipeline_script=f"{data_dir}/run.sh",
        run_id=run_id
    )
    await evenet_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_STOPED,analysis_)


@container_controller.post("/stop-container-by-run-id/{run_id}")
@inject
async def run_container_by_run_id(
    run_id,
    # executor: JobExecutor = Depends(get_executor_dep),
    evenet_bus:EventBus = Depends(Provide[AppContainer.event_bus]) 
    ):
    settings = get_settings()

    data_dir = str(settings.DATA_DIR)

    analysis_ =  AnalysisExecuterModal(
        analysis_id=run_id,
        container_id=run_id,
        output_dir=data_dir,
        pipeline_script=f"{data_dir}/run.sh",
        run_id=run_id
    )
    await evenet_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_STOPED,analysis_)


@container_controller.post("/pull-image/{container_id}")
@inject
async def page_container(container_id,
                         job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    with get_engine().begin() as conn:
        find_container = container_service.find_container_by_id(conn,container_id)
        

        asyncio.create_task(job_executor.pull_image(container_id,find_container.image))
        container_service.update_container(conn,
                                            container_id,
                                            {"image_status":"pulling"})
    return "success"


@container_controller.post("/list-container-key")
@inject
async def  list_container_key(query:ListContainerQuery,
                               job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    with get_engine().begin() as conn:
        find_container=  container_service.list_container_key(conn,query)
        find_container = [dict(item) for item in find_container]
        items = await format_container(job_executor, find_container)
        
    return items




@container_controller.get("/inspect/{container_id}", tags=['container'])
@inject
async def get_inspect(container_id: str,
                        run_type:Optional[str]=None,
                         job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    run_id = container_id
    if  run_type:
        run_id = f"{run_type}-{container_id}"
    container_attr = await job_executor.get_container_attr(run_id)
    if container_attr:
        return container_attr
    else:
        raise HTTPException(status_code=404, detail=f"Container {run_id} not found or not running")




@container_controller.get("/image-inspect/{image_name}", tags=['container'])
@inject
async def get_image_inspect(image_name: str,
                         job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    image_attr = await job_executor.get_image_attr(image_name)
    if image_attr:
        return image_attr
    else:
        raise HTTPException(status_code=404, detail=f"Image {image_name} not found")
    


@container_controller.get("/list-running-container", tags=['container'])
@inject
async def list_running_container(
                        force: Optional[bool] = False,
                         job_executor:JobExecutor = Depends(Provide[AppContainer.job_executor_selector]) ):
    if force:
        return await job_executor.refresh_list_running()
    else:
        return await job_executor.list_running()
    



def get_ids(values):
    if isinstance(values, dict):
        if "sample" in values:
            return values['sample']
        elif "value" in values and  "label" in values:
            return values['value']
        else:
            return []
    return values

def get_group(values):
    if isinstance(values, dict):
        if "group" in  values:
            if len(values["group"]) ==1:
                return values["group"][0]
            elif len(values["group"]) >1:
                return "-".join(values["group"])
    return "-"



from operator import and_, or_

from shortuuid import uuid
from brave.microbe.schemas.entity import PageEntity
from brave.microbe.models.core import t_association,t_mesh,t_study
from sqlalchemy import select,func
from sqlalchemy.engine import Connection
from brave.microbe.service import graph_service
from brave.microbe.service import mesh_service,study_service


def page(conn: Connection, query: PageEntity):
    mesh_subject = t_mesh.alias("mesh_subject")
    mesh_object = t_mesh.alias("mesh_object")
    mesh_observed = t_mesh.alias("mesh_observed")
    mesh_study = t_mesh.alias("mesh_study")
    association = t_association.alias("assoc")  # 关系表

    stmt = (
        select(
            association,
            mesh_subject.c.entity_name.label("subject_name"),
            mesh_object.c.entity_name.label("object_name"),
            mesh_observed.c.entity_name.label("observed_name"),
            mesh_study.c.entity_name.label("study_name")
        )
        .select_from(association)
        .outerjoin(mesh_subject, mesh_subject.c.entity_id == association.c.subject_id)
        .outerjoin(mesh_object, mesh_object.c.entity_id == association.c.object_id)
        .outerjoin(mesh_observed, mesh_observed.c.entity_id == association.c.observed_id)  # 正确字段
        .outerjoin(mesh_study, mesh_study.c.entity_id == association.c.study_id)
    )
    conditions = []
  
    if query.keywords:
        keyword_pattern = f"%{query.keywords.strip()}%"
        conditions.append(
             t_association.c.entity_name.ilike(keyword_pattern)
        )
    
    
    if conditions:  # 只有有条件时才加 where
        conditions = and_(*conditions) if len(conditions) > 1 else conditions[0]
        stmt = stmt.where(conditions)
        count_stmt = select(func.count()).select_from(t_association).where(conditions)
    else:
        count_stmt = select(func.count()).select_from(t_association)
    if query.page_size != -1:
        stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    stmt = stmt.order_by(association.c.id.desc())
    result  = conn.execute(stmt).mappings().all()
    total = conn.execute(count_stmt).scalar()

    formatted_result = []
    for row in result:
        new_row = dict(row)  # 转成普通字典
        for field in ["created_at", "updated_at"]:
            if field in new_row and new_row[field]:
                new_row[field] = new_row[field].strftime("%Y-%m-%d %H:%M:%S")
        formatted_result.append(new_row)

    return {
        "items": formatted_result ,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }

def find_by_id(conn: Connection, entity_id: str):
    stmt = t_association.select().where(t_association.c.entity_id ==entity_id)
    find_disease = conn.execute(stmt).mappings().first()
    return find_disease

def update(conn: Connection, entity_id: str, updateData: dict):
    stmt = t_association.update().where(t_association.c.entity_id == entity_id).values(updateData)
    conn.execute(stmt)

def add(conn: Connection, data: dict):
    data['entity_id'] = str(uuid())
    stmt = t_association.insert().values(data)
    conn.execute(stmt)
    return data['entity_id']
    # for item in data.items():
    #     print(item)
    #     pass
    # graph_service.create_node_relation()
    


def find_by_keywords(conn: Connection, keywords: str):
    keyword_pattern = f"%{keywords.strip()}%"
    stmt = t_association.select().where(t_association.c.entity_name.ilike(keyword_pattern)).limit(10)
    find_disease = conn.execute(stmt).mappings().all()
    return find_disease

def details_by_id(conn: Connection, entity_id: str):
    stmt = t_association.select().where(t_association.c.entity_id == entity_id)
    details_disease = conn.execute(stmt).mappings().first()
    return details_disease

def imports(conn: Connection, records: list, batch_size: int = 1000):
    inserted = 0
    with conn.begin():  # type: Connection
        for i in range(0, len(records), batch_size):
            batch = records[i:i + batch_size]
            stmt = t_association.insert()
            conn.execute(stmt, batch)
            inserted += len(batch)
            print(f"Inserted {inserted} records...")

    return {"message": f"导入完成，共导入 {inserted} 行"}

def delete_by_id(conn: Connection, entity_id: str):
    stmt = t_association.delete().where(t_association.c.entity_id == entity_id)
    conn.execute(stmt)


def get_entity(conn: Connection,association:dict):
    subject_id = association.get("subject_id") 
    object_id = association.get("object_id") 
    observed_id = association.get("observed_id")
    predicate = association.get("predicate") 
    study_id = association.get("study_id") 
    

    subject = mesh_service.find_by_id(conn,subject_id)


    object = mesh_service.find_by_id(conn,object_id)
    
    observed = mesh_service.find_by_id(conn,observed_id)

    study = study_service.find_study_by_id(conn,study_id)


    entity = {
        "subject":subject,
        "object":object,
    }
    relationship = {
        "predicate": predicate,
        "association_id": association.get("entity_id"), 
        "effect": association.get("effect"),
        "study":study.get("entity_id"),
        "study_name":study.get("entity_name"),
        "observed_id":observed.get("entity_id"),
        "observed_name":observed.get("entity_name")
    }



    # result.update({"association":association})
    
    return entity,relationship



from fastapi import APIRouter
from Bio.KEGG.KGML.KGML_parser import read
import requests
import os
# from lxml import etree

kegg_api = APIRouter(prefix="/kegg")


def download_kgml(pathway_id: str) -> str:
    """下载 KEGG KGML 文件"""
    filename = f"{pathway_id}.kgml"
    if not os.path.exists(filename):
        url = f"http://rest.kegg.jp/get/{pathway_id}/kgml"
        resp = requests.get(url)
        if resp.status_code != 200:
            raise ValueError(f"无法获取 {pathway_id} 的 KGML 文件")
        with open(filename, "w", encoding="utf-8") as f:
            f.write(resp.text)
    return filename


@kegg_api.get("/pathway/{pathway_id}/markers")
def get_markers(pathway_id: str):
    up = {"C04549","5337"}
    down = {"C00024"}
    kgml_file = download_kgml(pathway_id)

    with open(kgml_file) as f:
        p = read(f)

    markers = []
    for e in p.entries.values():
        if e.type == "compound" or e.type == "gene":
            for n in e.name.split():
                cid = n.split(":")[1]
                if cid in up:
                    graphics = e.graphics[0]
                    markers.append({"id": cid, "type":graphics.type,"radius":graphics.width/2,"x": graphics.x, "y": graphics.y, "width": graphics.width, "height": graphics.height ,"status": "up"})
                elif cid in down:
                    markers.append({"id": cid, "x": e.graphics.x, "y": e.graphics.y, "status": "down"})

    return markers



@kegg_api.get("/kgml/{pathway_id}")
def parse_kgml(pathway_id: str):
    file_path = download_kgml(pathway_id)
    # file_path = f"/ssd1/wy/workspace2/nextflow-fastapi/hsa04144.kgml"  # 存放KGML文件
    tree = etree.parse(file_path)
    root = tree.getroot()

    image = root.attrib.get("image")
    title = root.attrib.get("title")

    entries = []
    for entry in root.findall("entry"):
        eid = entry.attrib.get("id")
        name = entry.attrib.get("name")
        type_ = entry.attrib.get("type")
        link = entry.attrib.get("link")
        name = name.split()
        if type_=="compound":
            name = [item.replace("cpd:","") for item in name]
        g = entry.find("graphics")
        if g is None:
            continue
        if g.attrib.get("type")=="line":
            continue
        x = float(g.attrib.get("x"))
        y = float(g.attrib.get("y"))
        w = float(g.attrib.get("width", 0))
        h = float(g.attrib.get("height", 0))
        label = g.attrib.get("name", "")
        entries.append({
            "id": eid,
            "name": name,
            "type": type_,
            "link": link,
            "x": x,
            "y": y,
            "width": w,
            "height": h,
            "label": label,
        })

    return {
        "title": title,
        "image": image,
        "entries": entries
    }


from fastapi import APIRouter, Request, HTTPException,Depends   
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
from brave.api.service.sse_service import SSESessionService
import json
from dependency_injector.wiring import inject, Provide
from brave.app_container import AppContainer
sseController = APIRouter()



# # 全局生产者队列（数据源）
# global_queue = asyncio.Queue()

# # 所有客户端连接（用于广播）
# connected_clients = set()

# # 消费全局队列并广播给每个客户端
# async def broadcast_loop():
#     current_loop = asyncio.get_event_loop()
#     print(f"broadcast_loop 事件循环：{current_loop}")
#     while True:
#         msg = await global_queue.get()
#         print(f"广播消息{msg} 客户端数量:{len(connected_clients)}")
#         # 广播给所有客户端
#         for q in connected_clients.copy():
#             await q.put(msg)

# # SSE 消息生成器，每个连接一个队列
# async def event_generator(request: Request, client_queue: asyncio.Queue):
#     try:
#         while True:
#             # print(f"is_disconnected: {request.is_disconnected()}")
#             if await request.is_disconnected():
#                 print("请求关闭!")
#                 break
#             try:
#                 msg = await asyncio.wait_for(client_queue.get(), timeout=10)
#                 print(f"产生消息{msg}!")
#                 yield f"data: {msg}\n\n"
#             except asyncio.TimeoutError:
#                 yield ": keep-alive\n\n"
#     except asyncio.CancelledError:
#         print("连接被取消")
#     finally:
#         print("finally请求关闭!")
#         connected_clients.discard(client_queue)

# @sseController.get("/sse")
# async def sse(request: Request):
#     q = asyncio.Queue()
#     connected_clients.add(q)
#     return StreamingResponse(event_generator(request, q), media_type="text/event-stream")

# @sseController.get("/send")
# async def send_message(msg: str):
#     await global_queue.put(msg)
#     return {"status": "queued", "message": msg}




# # 生产者示例：每隔5秒往队列里放一个数据
# async def  producer():
#     i = 1
#     while True:
#         await asyncio.sleep(10)
#         print(f"📦 当前线程：{threading.current_thread().name}, 消息 {i}")
#         await global_queue.put(f"消息 {i}")
#         i += 1

# threading.Thread(target=producer, daemon=True).start()

# from brave.api.service.sse_service import sse_service  # 从 service.py 导入


# @sseController.get("/sse")
# async def sse(request: Request):
#     q = asyncio.Queue()
#     sse_service = request.app.state.sse_service  # 从 app.state 获取实例
#     sse_service.add_client(q)
#     return StreamingResponse(sse_service.event_generator(request, q), media_type="text/event-stream")


# @sseController.get("/sse-group")
# async def sse_group(request: Request,group="default"):
#     q = asyncio.Queue()
#     manager: AppManager = request.app.state.manager  # 从 app.state 获取实例
#     if manager.sse_service is None:
#         raise HTTPException(status_code=500, detail="SSE服务未初始化")
        
#     manager.sse_service.add_client(q,group)
#     return StreamingResponse(manager.sse_service.event_generator(request, q,group), media_type="text/event-stream")


# @sseController.get("/send")
# async def send_message(msg: str, request: Request):
#     manager: AppManager = request.app.state.manager  # 从 app.state 获取实例
#     if manager.sse_service is None:
#         raise HTTPException(status_code=500, detail="SSE服务未初始化")
#     await manager.sse_service.push_message({"group":"default","data":msg})
#     return {"status": "queued", "message": msg}

@sseController.get("/send-test")
@inject
async def send_message2(sse_service:SSESessionService = Depends(Provide[AppContainer.sse_service])  ):
    await sse_service.push_message({"group":"default","data":json.dumps({"msgType":"test","msg":"hello"})})
    return {"message": "success"}


from typing import Callable, Awaitable, Dict, Union, Coroutine
import asyncio
from collections import defaultdict
from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.event import WorkflowEvent
from brave.api.schemas.analysis import Analysis
from brave.api.service import analysis_service
from brave.api.config.db import get_engine

Callback = Union[Callable[[Analysis,dict], None],Callable[[Analysis,dict], Coroutine]]

class WorkflowEventRouter(BaseEventRouter[WorkflowEvent,Callback]):
    def __init__(self):
        super().__init__()
        self._analysis_cache: dict[str, Analysis] = {}

    def invalidate_cache(self, analysis_id: str):
        """清除缓存"""
        self._analysis_cache.pop(analysis_id, None)

    async def get_analysis(self, analysis_id: str) -> Analysis:
        if analysis_id not in self._analysis_cache:
            with get_engine().begin() as conn:
                analysis =  analysis_service.find_analysis_by_id(conn,analysis_id)
                if analysis is None:
                    raise ValueError(f"Analysis with id {analysis_id} not found")
                analysis_dict = dict(analysis)  # Convert Row to dict if needed
                analysis = Analysis(**analysis_dict)
                self._analysis_cache[analysis_id] = analysis
        return self._analysis_cache[analysis_id]

    async def dispatch(self,event:WorkflowEvent, analysis_id:str,msg: dict):
        # event = msg.get("workflow_event")
        # if not event:
        #     print("[EventRouter] No 'workflow_event' in message", msg)
        #     return
        analysis = await self.get_analysis(analysis_id)
        handlers = self._handlers.get(event)
        if handlers:
            for handler in handlers:
                if asyncio.iscoroutinefunction(handler):
                    await handler(analysis,msg)
                else:
                    handler(analysis,msg)
        else:
            print(f"[WorkflowEventRouter] No handler for event '{event}'")

        if event == WorkflowEvent.ON_FLOW_COMPLETE:
            self.invalidate_cache(analysis_id)

# @lru_cache(maxsize=1)
# def get_router():
#     return WorkflowEventRouter()


import xml.etree.ElementTree as ET
import json
import pandas as pd
def get_json(file_path):
    # 解析 XML 文件
    tree = ET.parse("/ssd1/wy/workspace2/nextflow-fastapi/desc2025.xml")
    root = tree.getroot()

    mesh_json = []

    for descriptor in root.findall("DescriptorRecord"):
        # DescriptorUI
        ui_elem = descriptor.find("DescriptorUI")
        descriptor_ui = ui_elem.text if ui_elem is not None else None

        # DescriptorName
        name_elem = descriptor.find("DescriptorName/String")
        descriptor_name = name_elem.text if name_elem is not None else None
    # RegistryNumberList（新增）
        registry_numbers = []
        concept_elem = descriptor.find("ConceptList/Concept")
        if concept_elem is not None:
            reg_list_elem = concept_elem.find("RegistryNumberList")
            if reg_list_elem is not None:
                registry_numbers = [rn.text for rn in reg_list_elem.findall("RegistryNumber")]
        # TreeNumberList
        tree_numbers = []
        tree_list_elem = descriptor.find("TreeNumberList")
        if tree_list_elem is not None:
            tree_numbers = [tn.text for tn in tree_list_elem.findall("TreeNumber")]

        # PublicMeSHNote
        note_elem = descriptor.find("PublicMeSHNote")
        public_note = note_elem.text if note_elem is not None else None

        # PreviousIndexingList
        prev_indexing = []
        prev_list_elem = descriptor.find("PreviousIndexingList")
        if prev_list_elem is not None:
            prev_indexing = [pi.text for pi in prev_list_elem.findall("PreviousIndexing")]

        # 构建 JSON 对象
        descriptor_dict = {
            "DescriptorUI": descriptor_ui,
            "DescriptorName": descriptor_name,
            "TreeNumberList": tree_numbers,
            "PublicMeSHNote": public_note,
            "registry_numbers":registry_numbers,
            "PreviousIndexingList": prev_indexing
        }

        mesh_json.append(descriptor_dict)
    return mesh_json




import json
from brave.api.core import evenet_bus
from brave.api.core.evenet_bus import EventBus
from brave.api.core.routers_name import RoutersName
from brave.api.schemas.analysis import Analysis
from brave.api.schemas.analysis_result import AnalysisResult, AnalysisResultParseModal
from brave.api.service import analysis_result_service
from brave.api.service import analysis_service
from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.api.config.db import get_engine
from brave.api.core.event import AnalysisResultEvent
from collections import defaultdict

import asyncio
from typing import  Dict


class ResultParse:
    def __init__(self,analysis_id,evenet_bus:EventBus) -> None:
        self.analysis_id = analysis_id
        self.event_bus = evenet_bus
        self.analysis_result_parse_service = AnalysisResultParse()
        self.analysis_locks: Dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)

    async def parse(self,save:bool=True):
        async with self.analysis_locks[self.analysis_id]:
            with get_engine().begin() as conn:
                params = analysis_service.get_parse_analysis_result_params(conn,self.analysis_id)
                result_list,result_dict = analysis_service.execute_parse(**params)
                # params,result_list,result_dict = self.analysis_result_parse_service.parse_analysis_result(conn,self.analysis_id,True)
                print("result_list",json.dumps(result_list,indent=4))
                for item in result_list:    
                    result = self.analysis_result_parse_service.find_analysis_result_exist(conn,self.analysis_id,item['component_id'],item['file_name'],item['project'],True)
                    if not result:
                        find_sample = self.analysis_result_parse_service.find_by_sample_name_and_project(conn,item['file_name'],item['project'])
                        if find_sample:
                            item['sample_id'] = find_sample['sample_id']
                        if save:
                            analysis_result_service.add_analysis_result(conn,item)
                        analsyis = dict(params['analysis'])
                        analsyis = Analysis(**analsyis)
                        analysis_result = AnalysisResultParseModal(**item)
                        await self.event_bus.dispatch(RoutersName.ANALYSIS_RESULT_ROUTER, AnalysisResultEvent.ON_ANALYSIS_RESULT_ADD, analsyis, analysis_result)
                        # await self.listener_files_service.execute_listener("analysis_result_add",{
                        #     "analysis":params['analysis'],
                        #     "analysis_result":item,
                        #     "sse_service":self.sse_service
                        # })
                        # data = json.dumps({
                        #     "msg":f"分析{analysis_id}，文件{item['file_name']}保存成功!",
                        #     "component_id":item['component_id'],
                        #     "msgType":"analysis_result"
                        # })
                        # msg = {"group": "default", "data": data}
                        # await self.sse_service.push_message(msg)
                    else:
                        if item['analysis_result_hash']!= result['analysis_result_hash']:
                            if save:
                                analysis_result_service.update_analysis_result(conn,result.id,item)
                            analsyis = dict(params['analysis'])
                            result = dict(result)
                            analsyis = Analysis({**analsyis,**result})
                            analysis_result = AnalysisResultParseModal(**item)
                            await self.event_bus.dispatch(RoutersName.ANALYSIS_RESULT_ROUTER, AnalysisResultEvent.ON_ANALYSIS_RESULT_UPDATE, analsyis, analysis_result)
        
        return result_list,result_dict,params


import asyncio
import os
import json
from typing import Optional

from brave.api.config.config import get_settings
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
# from brave.api.service.file_watcher_service import FileWatcher
from brave.api.handlers import analysis_executer
from brave.api.service.listener_files_service import get_listener_files_service
from brave.api.service.process_monitor_service import ProcessMonitor
from brave.api.service.sse_service import SSESessionService
from brave.api.config.db import get_engine
from brave.api.service import namespace_service, project_service
from brave.api.ingress.manager import IngressManager
from brave.api.handlers import workflow_events,analysis_result   
from brave.api.core.workflow_queue import WorkflowQueueManager
from dependency_injector.wiring import inject, Provide
from brave.app_container import AppContainer
from brave.api.core.routers.ingress_event_router import IngressEventRouter
from brave.api.core.event import IngressEvent
from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.api.service.listener_files_service import ListenerFilesService
from brave.api.core.heartbeat import process_heartbeat
from brave.api.core.routers.watch_file_event_router import WatchFileEvenetRouter
from brave.api.service.file_watcher_service import FileWatcherService
from brave.api.core.event import WatchFileEvent
from brave.api.core.event import WorkflowEvent
import  brave.api.service.analysis_service as analysis_service
from brave.api.executor.base import JobExecutor
from brave.api.executor.local_executor import LocalExecutor
from py2neo import Graph

class AppManager:
    @inject
    def __init__(
        self, 
        workflow_queue_manager: WorkflowQueueManager = Provide[AppContainer.workflow_queue_manager],
        ingress_manager: IngressManager = Provide[AppContainer.ingress_manager],
        ingress_event_router: IngressEventRouter = Provide[AppContainer.ingress_event_router],
        sse_service: SSESessionService = Provide[AppContainer.sse_service],
        analysis_result_parse_service: AnalysisResultParse = Provide[AppContainer.analysis_result_parse_service],
        listener_files_service: ListenerFilesService = Provide[AppContainer.listener_files_service],
        workflow_event_router:WorkflowEventRouter=Provide[AppContainer.workflow_event_router],
        watchfile_event_router:WatchFileEvenetRouter=Provide[AppContainer.watchfile_event_router],
        job_executor:JobExecutor=Provide[AppContainer.job_executor_selector],
        config = Provide[AppContainer.config]   
        
        ):
        self.config = config
        self.graph: Graph | None = None
        self.workflow_queue_manager = workflow_queue_manager
        self.ingress_event_router = ingress_event_router
        self.ingress_manager = ingress_manager
        self.sse_service = sse_service
        self.analysis_result_parse_service = analysis_result_parse_service
        self.listener_files_service = listener_files_service
        self.workflow_event_router = workflow_event_router
        self.watchfile_event_router = watchfile_event_router
        self.job_executor = job_executor
        self.config = config
        self.tasks = []
        # 预先声明属性，后面启动时赋值
        self.file_watcher = None
        self.process_monitor = None
        self.wes = None
        


    async def start(self):
        settings = get_settings()
        watch_path = f"{settings.BASE_DIR}/monitor"
        if not os.path.exists(watch_path):
            os.makedirs(watch_path)

        # self.graph = Graph(
        #     self.settings.bolt_url,
        #     auth=(self.settings.user, self.settings.password)
        # )

        # self.graph  = Graph("bolt://localhost:7687", auth=("neo4j", "password"))

        self.file_watcher_service = FileWatcherService(
            watch_path=watch_path,
            watchfile_event_router=self.watchfile_event_router
        ) 
        self.process_monitor = ProcessMonitor(
            sse_service=self.sse_service,
            analysis_result_parse_service=self.analysis_result_parse_service,
            listener_files_service=self.listener_files_service
        )
      
        # register http ingress
        # self.ingress_manager.register_http(self.app)
        self.tasks.append(asyncio.create_task(self.ingress_manager.start()))

        # register handler for ingress event
        self.ingress_event_router.register_handler(IngressEvent.HEARTBEAT, process_heartbeat)

        # register  handler for workflow  queue
        # self.ingress_event_router.register_handler(IngressEvent.NEXTFLOW_EXECUTOR_EVENT, self.workflow_queue_manager.dispatch)
        # self.workflow_queue_manager.register_subscriber(WorkflowEvent.ON_FLOW_BEGIN, self.workflow_event_router.dispatch)
        # self.tasks.append(asyncio.create_task(self.workflow_queue_manager.cleanup_loop()))



        async def workflow_evnet_dispatch(msg:dict):
            try:
                event = WorkflowEvent(msg.get("workflow_event"))
            except ValueError:
                event = msg.get("workflow_event")
                print(f"[WorkflowEventRouter] Unknown event type '{event}'", msg)
                return
            analysis_id = msg.get("analysis_id")
            if analysis_id:
                await self.workflow_event_router.dispatch(event,analysis_id,msg)

        self.ingress_event_router.register_handler(IngressEvent.NEXTFLOW_EXECUTOR_EVENT,  workflow_evnet_dispatch)





        # async def push_default_message(analysis_id:str,msg:dict):
        #     await self.sse_service.push_message({"group": "default", "data": json.dumps(msg)})


        
        # #  sse_service.push_message
        # self.workflow_event_router.register_handler(WorkflowEvent.ON_FLOW_BEGIN,  push_default_message)
        # # self.workflow_event_router.register_handler(WorkflowEvent.ON_FILE_PUBLISH,  self.sse_service.push_message_default)
        # self.workflow_event_router.register_handler(WorkflowEvent.ON_PROCESS_COMPLETE,  push_default_message)
        # self.workflow_event_router.register_handler(WorkflowEvent.ON_FLOW_COMPLETE,  push_default_message)
        # self.workflow_event_router.register_handler(WorkflowEvent.ON_JOB_SUBMITTED,  push_default_message)

     
        # self.workflow_event_router.register_handler(WorkflowEvent.ON_FLOW_COMPLETE, finished_analysis_handler)

        # self.workflow_queue_manager.register_subscriber("", subscriber)

        async def push_file_watch_message(msg:dict):
            await self.sse_service.push_message({"group": "default", "data": json.dumps(msg)})

        # self.watchfile_event_router.register_handler(WatchFileEvent.WORKFLOW_LOG,push_file_watch_message)
        # self.watchfile_event_router.register_handler(WatchFileEvent.TRACE_LOG,  push_file_watch_message)

        # setup_handlers()
        analysis_executer.setup_handlers()
        workflow_events.setup_handlers()
        analysis_result.setup_handlers()
        self.tasks.append(asyncio.create_task(self.sse_service.broadcast_loop()))



        # self.tasks.append(asyncio.create_task(self.analysis_result_parse_service.auto_save_analysis_result()))
        self.tasks.append(asyncio.create_task(self.file_watcher_service.watch_folder()))
        self.tasks.append(asyncio.create_task(self.process_monitor.startup_process_event()))
        # 挂载到 app.state，方便别处访问
        # self.app.state.manager = self

        # init db
        with get_engine().begin() as conn: 
            await namespace_service.init_db(conn)
            await project_service.init_db(conn)

        

    async def stop(self):
        for task in self.tasks:
            task.cancel()
        await asyncio.gather(*self.tasks, return_exceptions=True)
        self.graph = None
        print("[AppManager] All background tasks stopped")



import os
import asyncio
import json
from .interfaces.base_ingress import BaseMessageIngress
from brave.api.core.routers.ingress_event_router import IngressEventRouter
from brave.api.core.event import IngressEvent
class UDSStreamIngress(BaseMessageIngress):
    def __init__(self, path, router:IngressEventRouter):
        self.path = path
        self.router = router

    async def start(self):
        if os.path.exists(self.path):
            os.remove(self.path)
        server = await asyncio.start_unix_server(self._handle, path=self.path)
        print(f"[UDS-STREAM] Listening at {self.path}")
        async with server:
            await server.serve_forever()

    async def _handle(self, reader, writer):
        try:
            while line := await reader.readline():
                try:
                    msg = json.loads(line.decode().strip())
                    
                    evnet_str = msg.get("ingress_event")
                    event = IngressEvent(evnet_str)
                    if not event:
                        event = IngressEvent.NEXTFLOW_EXECUTOR_EVENT
                        print(f"[UDS-STREAM] Unknown event type '{event}'", msg)
                        return
                    data = msg.get("data")
                    await self.router.dispatch(event,data)
                except Exception as e:
                    print(f"[UDS-STREAM] Message error: {e}")
        except Exception as e:
            print(f"[UDS-STREAM] Client error: {e}")
        finally:
            writer.close()
            await writer.wait_closed()



from typing import Optional,Any
from pydantic import BaseModel
from typing import Dict
class PageEntity(BaseModel):
    page_number: Optional[int]=1
    page_size: Optional[int]=10
    parent_id: Optional[str]=None
    category: Optional[Any]=None
    # major_category: Optional[str]=None
    keywords: Optional[str]=None
    is_research: Optional[bool]=None
    locale:Optional[str]="en_US"
    registry_join: Optional[str]=None




class Entity(BaseModel):
    label: str                 # 节点标签，比如 "Taxonomy", "Study", "Disease"
    entity_id: str             # 唯一ID
    properties: Dict[str, Any] # 其他属性，可选

class RelationshipRequest(BaseModel):
    from_entity: Entity
    to_entity: Entity
    relation_type: str         # 关系类型，比如 "ASSOCIATED_WITH"

class GraphQuery(BaseModel):
    label: Optional[str] = None
    keyword: Optional[str] = None
    entity_id: Optional[str] = None
    nodes:Optional[list[str]] = None
    nodes_dict: Optional[Dict[str, list[str]]] = None
    nodes_dict_condition: Optional[str] = "OR"  # "AND" 或 "OR"
    order_by: Optional[str] = "taxonomy"  # 排序字段
    order_metric: Optional[str] = "study"  # study
    collect_association_study: Optional[bool]= False
    locale:Optional[str]="en_US"


class GraphQueryV2(BaseModel):
    entity_id: Optional[str]=None
    entity_type: Optional[str]=None  # taxonomy / disease / intervention / microbe / study
    keyword: Optional[str]=None
    depth: int = 1  # 关系深度，可控制查询范围
    relation_types: Optional[list[str]]=None  # 指定需要的关系类型


class NodeQuery(BaseModel):
    label: str                       # 节点 label
    keyword: Optional[str] = None    # 可选关键字
    page: int = 1                    # 页码，从 1 开始
    page_size: int = 20              # 每页数量


class DetailsNodeQuery(BaseModel):
    nodes: Optional[list[str]] = []  # 需要查询的关联节点类型，如 taxonomy, disease, diet_and_food, study


class AddMeshNode(BaseModel):
    category: Optional[str]  
    entity_name: str
    entity_name_zh: Optional[str] = None
    entity_type: Optional[str] = None
    tags: Optional[str] = None
    short_name: Optional[str] = None    
    describe: Optional[str] = None

    parent_tree: Optional[str] = None  # 可以为 None





# brave/container.py
from anyio import Event
from dependency_injector import containers, providers
from brave.api.core.evenet_bus import EventBus
from brave.api.core.routers.analysis_result_router import AnalysisResultRouter
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
from brave.api.core.workflow_queue import WorkflowQueueManager
from brave.api.core.pubsub import PubSubManager
from brave.api.core.routers.ingress_event_router import IngressEventRouter
from brave.api.ingress.manager import IngressManager
from brave.api.core.workflow_sse import WorkflowSSEManager
from brave.api.service.result_parse.analysis_manage import AnalysisManage
from brave.api.service.result_parse.nextflow_analysis import NextflowAnalysis
from brave.api.service.result_parse.script_analysis import ScriptAnalysis
from brave.api.service.sse_service import SSESessionService
from brave.api.service.analysis_result_parse import AnalysisResultParse
from brave.api.service.listener_files_service import ListenerFilesService
from brave.api.ingress.factory import IngressMode
from brave.api.core.routers.watch_file_event_router import WatchFileEvenetRouter
from brave.api.core.routers.analysis_executer_router import AnalysisExecutorRouter
# from brave.api.executor.factory import get_executor
from brave.api.executor.docker_excutor import DockerExecutor
from brave.api.executor.k8s_executor import K8sExecutor
from brave.api.executor.slurm_executor import SlurmExecutor
from brave.api.executor.local_executor import LocalExecutor
class AppContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    # Core services
    wiring_config = containers.WiringConfiguration(modules=[".api",".app_manager",".api.routes"])
    pubsub_manager = providers.Singleton(PubSubManager)
    workflow_event_router = providers.Singleton(WorkflowEventRouter)
    watchfile_event_router = providers.Singleton(WatchFileEvenetRouter)
    analysis_executer_router = providers.Singleton(AnalysisExecutorRouter)
    analysis_result_router = providers.Singleton(AnalysisResultRouter)


    
    

    workflow_queue_manager = providers.Singleton(
        WorkflowQueueManager, 
        pubsub=pubsub_manager,
        workflow_event_router=workflow_event_router
    )
    
    # Create adapter for WorkflowEventSystem
    ingress_event_router = providers.Singleton(IngressEventRouter)

    ingress_manager = providers.Singleton(
        IngressManager,
        event_mode=IngressMode.STREAM,
        uds_path="/tmp/brave.sock",
        ingress_event_router=ingress_event_router
    )

    # Workflow event system
    # workflow_queue_manager = providers.Singleton(
    #     WorkflowQueueManager,
    #     pubsub=pubsub_manager,
    #     workflow_event_router=workflow_event_router
    # )
    sse_service = providers.Singleton(SSESessionService)
    workflow_sse_manager = providers.Singleton(WorkflowSSEManager, workflow_queue_manager=workflow_queue_manager)
    
    event_bus = providers.Singleton(
        EventBus,
    )
    nextflow_analysis = providers.Singleton(NextflowAnalysis,event_bus=event_bus)
    script_analysis = providers.Singleton(ScriptAnalysis,event_bus=event_bus)
    analysis_controller_selector = providers.Selector(
        lambda type: type,  
        script=script_analysis,
        nextflow=nextflow_analysis
    )
    
    
    listener_files_service = providers.Singleton(ListenerFilesService)

    analysis_result_parse_service = providers.Singleton(
        AnalysisResultParse
  
    )
    result_parse_manage= providers.Singleton(AnalysisManage, event_bus=event_bus)
    
 
    docker_executor = providers.Singleton(DockerExecutor,event_bus=event_bus)
    k8s_executor = providers.Singleton(K8sExecutor,event_bus=event_bus)
    slurm_executor = providers.Singleton(SlurmExecutor,event_bus=event_bus)
    local_executor = providers.Singleton(LocalExecutor,event_bus=event_bus)
    job_executor_selector = providers.Selector(
        config.executer_type,
        local=local_executor,
        docker=docker_executor,
        k8s=k8s_executor,
        slurm=slurm_executor
    )
        
    # job_executor  =providers.Singleton(get_executor,"local")

    # App manager with injected dependencies
    # app_manager = providers.Singleton(
    #     AppManager,
    #     app=providers.Dependency(instance_of=FastAPI),
    #     workflow_event_system=workflow_event_system
    # )
    




import asyncio
from typing import Generic, TypeVar

from brave.api.core.base_event_router import BaseEventRouter
from typing import Callable, Awaitable, Coroutine, Union

Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]


E = TypeVar("E")  # 事件类型（如 Enum）
# F = TypeVar("F")
class DirectDispatch(BaseEventRouter[E,Callback]):
    async def dispatch(self, event: E, msg: dict):
        handlers = self._handlers.get(event, [])
        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler(msg)
            else:
                handler(msg)



from typing import Optional
from pydantic import BaseModel

class PageContainerQuery(BaseModel):
    page_number: Optional[int]=1
    page_size: Optional[int]=10

class SaveContainer(BaseModel):
    name: Optional[str]
    image: Optional[str]
    container_id: Optional[str]=None
    description: Optional[str]=None
    envionment:Optional[str]=None
    command: Optional[str]=None
    port: Optional[str]=None
    labels: Optional[str]=None
    change_uid:Optional[bool]=True
    container_key:Optional[str]=None
    img: Optional[str]=None
    

class ListContainerQuery(BaseModel):
    container_key: Optional[list]=None




# container/k8s_backend.py
from kubernetes import client, config
from brave.api.executor.models import JobSpec
from brave.api.core.evenet_bus import EventBus
from .base import JobExecutor   
from brave.api.core.routers.workflow_event_router import WorkflowEventRouter
class K8sExecutor(JobExecutor):
    def __init__(self,event_bus:EventBus):
        self.event_bus = event_bus
        config.load_kube_config()
        self.api = client.CoreV1Api()
        self.batch_api = client.BatchV1Api()

    async def _do_submit_job(self, job_spec) -> str:
        # 这里应构造 Job YAML spec
        job = client.V1Job(...)  # 构建 Job 对象
        self.batch_api.create_namespaced_job(namespace="default", body=job)
        return job.metadata.name

    def get_logs(self, job_id: str) -> str:
        pod_list = self.api.list_namespaced_pod("default", label_selector=f"job-name={job_id}")
        pod_name = pod_list.items[0].metadata.name
        return self.api.read_namespaced_pod_log(pod_name, namespace="default")

    def stop_job(self, job_id: str) -> None:
        self.batch_api.delete_namespaced_job(name=job_id, namespace="default", propagation_policy='Foreground')



from .docker_excutor import DockerExecutor
from .k8s_executor import K8sExecutor
from .slurm_executor import SlurmExecutor
from .local_executor import LocalExecutor
from .base import JobExecutor
from fastapi import Query
def get_executor(platform: str) -> JobExecutor:
    # workflow_event_router = AppContainer.workflow_event_router()
    if platform == "docker":
        return DockerExecutor()
    elif platform == "k8s":
        return K8sExecutor()
    elif platform == "slurm":
        return SlurmExecutor()
    elif platform == "local":
        return LocalExecutor()
    else:
        raise ValueError(f"Unsupported backend: {platform}")


# def get_executor_dep(
#     platform: str = Query(default="local")  # 可从请求中传入，也支持配置默认值
# ) -> JobExecutor:
#     return get_executor(platform)




import asyncio
import json
from dependency_injector.wiring import inject
from fastapi import HTTPException
from brave.api.core.evenet_bus import EventBus
from brave.api.core.routers.analysis_executer_router import AnalysisExecutorRouter
from dependency_injector.wiring import inject, Provide
from brave.api.core.routers_name import RoutersName
from brave.api.executor.base import JobExecutor
from brave.api.schemas.analysis import AnalysisExecuterModal
from brave.api.service import analysis_service
from brave.app_container import AppContainer
from brave.api.core.event import WorkflowEvent
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.executor.models import JobSpec, LocalJobSpec, DockerJobSpec
from brave.api.service.sse_service import SSESessionService
from brave.api.service.result_parse.analysis_manage import AnalysisManage
from brave.api.executor.local_executor import LocalExecutor

@inject
def setup_handlers(
    evenet_bus:EventBus  = Provide[AppContainer.event_bus],
    router:AnalysisExecutorRouter  = Provide[AppContainer.analysis_executer_router],
    job_executor:JobExecutor = Provide[AppContainer.job_executor_selector],
    sse_service:SSESessionService = Provide[AppContainer.sse_service],
    result_parse_manage:AnalysisManage = Provide[AppContainer.result_parse_manage]):
    
    evenet_bus.register_router(RoutersName.ANALYSIS_EXECUTER_ROUTER,router)

    @router.on_event(AnalysisExecutorEvent.ON_ANALYSIS_SUBMITTED)
    async def on_analysis_submitted(payload:AnalysisExecuterModal):
        print(f"🚀 [on_analysis_submitted] {payload.analysis_id}")
        await job_executor.submit_job(payload)


    @router.on_event(AnalysisExecutorEvent.ON_ANALYSIS_STOPED)
    async def on_analysis_stoped(payload:AnalysisExecuterModal):
        print(f"🚀 [on_analysis_stoped] {payload.analysis_id}")
        await job_executor.remove_job(payload.run_id)
        # if payload.run_type =="job":
        #     asyncio.create_task(analysis_service.finished_analysis(payload.analysis_id,payload.run_type,"failed"))
        # else:
        #     asyncio.create_task(analysis_service.finished_analysis(payload.analysis_id,payload.run_type,"stoped"))

    
    @router.on_event(AnalysisExecutorEvent.ON_ANALYSIS_COMPLETE)
    async def on_analysis_complete(payload:AnalysisExecuterModal):
        print(f"🚀 [on_analysis_complete] {payload.analysis_id} {payload.run_type}")
        if payload.run_type =="job":
            asyncio.create_task(result_parse_manage.parse(payload.analysis_id))
            
            asyncio.create_task(analysis_service.finished_analysis(payload.analysis_id,payload.run_type,"finished"))
        elif payload.run_type =="server":
            asyncio.create_task(analysis_service.finished_analysis(payload.analysis_id,payload.run_type,"stopped"))
        # if payload.run_type !="retry":
        #     if payload.run_type =="job":
        #         asyncio.create_task(result_parse_manage.parse(payload.analysis_id))
        #         # await result_parse_manage.parse(payload.analysis_id)
            
        #     elif payload.run_type =="server":
        #         await analysis_service.update_ports(payload.analysis_id,None )
                
            

        await sse_service.push_message({"group": "default", "data": json.dumps({
            "analysis_id": payload.analysis_id,
            "event": "analysis_complete",
             "run_id": payload.run_id,
            "run_type":payload.run_type
            
            })})

    @router.on_event(AnalysisExecutorEvent.ON_ANALYSIS_STARTED)
    async def on_analysis_started(payload:AnalysisExecuterModal):
        print(f"🚀 [on_analysis_started] {payload.analysis_id} {payload.run_type}")
        if payload.run_type =="server":
            await analysis_service.update_url(payload.analysis_id,f"http://10.110.1.11:5003/container/{payload.analysis_id}")
        await sse_service.push_message({"group": "default", "data": json.dumps({
            "analysis_id": payload.analysis_id,
            "event": "analysis_started",
             "run_id": payload.run_id,
            "run_type":payload.run_type
            })})
  

    @router.on_event(AnalysisExecutorEvent.ON_ANALYSIS_FAILED)
    async def on_analysis_failed(payload:AnalysisExecuterModal):
        print(f"🚀 [on_analysis_failed] {payload.analysis_id}")
        if payload.run_type =="server":
            await analysis_service.update_url(payload.analysis_id,None )
        if payload.run_type !="retry":
            asyncio.create_task(analysis_service.finished_analysis(payload.analysis_id,payload.run_type,"failed"))
        await sse_service.push_message({"group": "default", "data": json.dumps({
            "analysis_id": payload.analysis_id,
            "event": "analysis_failed",
             "run_id": payload.run_id,
            "run_type":payload.run_type
            })})
 

    @router.on_event(AnalysisExecutorEvent.ON_CONTAINER_PULLED)
    async def on_container_pulled(payload:AnalysisExecuterModal):
        await sse_service.push_message({"group": "default", "data": json.dumps({
            "analysis_id": payload.analysis_id,
            "run_id": payload.run_id,
            "event": "container_pulled",
            "run_type":payload.run_type
        })})


from typing import Callable, Awaitable, Dict, Union, Coroutine
import asyncio
from collections import defaultdict

from pydantic import BaseModel

from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.schemas.analysis import AnalysisExecuterModal, AnalysisId
import brave.api.service.analysis_service as analysis_service
from brave.api.config.db import get_engine

Callback = Union[Callable[[AnalysisExecuterModal], None], Callable[[AnalysisExecuterModal], Coroutine]]

class AnalysisExecutorRouter(BaseEventRouter[AnalysisExecutorEvent,Callback]):
    async def dispatch(self, event: AnalysisExecutorEvent, payload: BaseModel):
        handlers = self._handlers.get(event, [])
        # expected_type = self._payload_types.get(event)
        # if  expected_type:
        #     if not isinstance(payload, expected_type):
        #         raise TypeError(f"[EventRouter] Expected payload type {expected_type}, got {type(payload)}")
        if payload.run_id is None:
            raise ValueError("run_id is required in payload")
      
        run_type = payload.run_id.split("-")[0]
        analysis_id = payload.run_id.replace(f"{run_type}-","")

        if event == AnalysisExecutorEvent.ON_ANALYSIS_COMPLETE or event == AnalysisExecutorEvent.ON_ANALYSIS_FAILED:    
            if isinstance(payload, AnalysisId):
                # if payload.run_type =="retry":
                #     payload = AnalysisExecuterModal(analysis_id=payload.analysis_id,run_type=payload.run_type)
                # else:
                with get_engine().begin() as conn:
                    
                    analysis =  analysis_service.find_analysis_by_id(conn,analysis_id)
                    if analysis:
                        payload = AnalysisExecuterModal(run_id=payload.run_id,**analysis)
                    else:
                       
                        payload = AnalysisExecuterModal(run_id=payload.run_id,analysis_id=analysis_id)

                # payload = AnalysisExecuterModal(analysis_id=payload.analysis_id)
        elif event == AnalysisExecutorEvent.ON_CONTAINER_PULLED:
            payload = AnalysisExecuterModal(run_id=payload.run_id, analysis_id=analysis_id)

            
        if   not isinstance(payload, AnalysisExecuterModal):
            raise TypeError(f"[EventRouter] Expected payload type {AnalysisExecuterModal}, got {type(payload)}")
    

        run_id = payload.run_id
        if run_id.startswith("job-"):
            payload.run_type = "job"
        elif run_id.startswith("server-"):
            payload.run_type = "server"
        elif run_id.startswith("retry-"):
            payload.run_type = "retry"
        else:
            raise ValueError(f"Invalid run_id format: {run_id}")
            

        if handlers:
            for handler in handlers:
                if asyncio.iscoroutinefunction(handler):
                    await handler(payload)
                else:
                    handler(payload)
        else:
            print(f"[AnalysisExecutorRouter] No handler for event '{event}'")
            
    
    # async def dispatch(self,event:AnalysisExecutorEvent, analysis_id:str,msg: dict):
    #     # event = msg.get("workflow_event")
    #     # if not event:
    #     #     print("[EventRouter] No 'workflow_event' in message", msg)
    #     #     return
    #     handlers = self._handlers.get(event)
    #     if handlers:
    #         for handler in handlers:
    #             if asyncio.iscoroutinefunction(handler):
    #                 await handler(analysis_id,msg)
    #             else:
    #                 handler(analysis_id,msg)
    #     else:
    #         print(f"[EventRouter] No handler for event '{event}'")
    # # def __init__(self):
    #     self._handlers: dict[WatchFileEvent, set[Callback]] = defaultdict(set)

    # def on_event(self, event: WatchFileEvent):
    #     def decorator(func: Callback):
    #         self._handlers[event].add(func)
    #         return func
    #     return decorator

    # def register_handler(self, event: WatchFileEvent, handler: Callback):
    #     self._handlers[event].add(handler)


    # async def dispatch(self, event:WatchFileEvent,msg: dict):
    #     # event = msg.get("file_type")
    #     # if not event:
    #     #     print("[WatchFileEvenetRouter] No 'file_type' in message", msg)
    #     #     return
    #     handlers = self._handlers.get(event)
    #     if handlers:
    #         for handler in handlers:
    #             if asyncio.iscoroutinefunction(handler):
    #                 await handler(msg)
    #             else:
    #                 handler(msg)
    #     else:
    #         print(f"[WatchFileEvenetRouter] No handler for event '{event}'")




from fastapi import APIRouter, Request
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
import time
import queue
import threading
sseController = APIRouter()


# 创建线程安全的队列
event_queue = queue.Queue()

# SSE 推送生成器
def event_generator():
    while True:
        try:
            # 阻塞等待新数据（最多等待1秒，避免无法关闭）
            data = event_queue.get(timeout=1)
            print(data)
            yield f"data: {data}\n\n"
        except queue.Empty:
            # 没有数据就继续等待
            continue


@sseController.get("/sse")
async def sse(request: Request):
    return StreamingResponse(event_generator(), media_type="text/event-stream")


# 生产者示例：每隔5秒往队列里放一个数据
def producer():
    i = 1
    while True:
        time.sleep(5)
        event_queue.put(f"消息 {i}")
        i += 1

# 启动生产线程
# threading.Thread(target=producer, daemon=True).start()


import asyncio
from typing import TypeVar, Generic, Callable, Coroutine, Union, Dict, Set
from collections import defaultdict, deque
from dataclasses import dataclass, field
from brave.api.core.base_event_router import BaseEventRouter

E = TypeVar("E")  # 泛型事件类型
Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]

class QueuedDispatch(BaseEventRouter[E,Callback]):
    def __init__(self, *args, maxlen: int = 100, delay: float =0, **kwargs):
        super().__init__(*args, **kwargs)
        self._queues: Dict[E, asyncio.Queue] = {}
        self._tasks: Dict[E, asyncio.Task] = {}
        self.maxlen = maxlen
        self.delay = delay

    async def dispatch(self, event: E, msg: dict):
        if event not in self._queues:
            self._queues[event] = asyncio.Queue(maxsize=self.maxlen)
            self._tasks[event] = asyncio.create_task(self._consume_worker(event))
        await self._queues[event].put(msg)

    async def _consume_worker(self, event: E):
        queue = self._queues[event]
        handlers = self._handlers.get(event, [])
        while True:
            msg = await queue.get()
            for handler in handlers:
                if asyncio.iscoroutinefunction(handler):
                    await handler(msg)
                else:
                    handler(msg)
            await asyncio.sleep(self.delay)



# from brave.api.core.pubsub import PubSubManager
# from brave.api.ingress.interfaces.base_ingress import BaseMessageIngress
# from .workflow_queue import WorkflowQueueManager
# from .uds_listener import UDSListener
# from .sse import SSEManager
# import asyncio
# from typing import Union, Optional

#                         [ workflow_queues ]
#                           ┌────────────┐
# [ UDS Producer ] ───→──→─▶ wf-123 → Queue ──┐
#                           └────────────┘    │
#                                             ▼
#                                    [ 消费 Task: while True get() ]
#                                             │
#                                             ▼
#                                    pubsub.publish(wf-123, msg)
#                                             │
#                 ┌───────────────────────────┼────────────────────────┐
#                 ▼                           ▼                        ▼
#       Log模块订阅器                  指标模块订阅器              告警模块订阅器



# [ Nextflow #1 ]──┐
# [ Nextflow #2 ]──┼──▶  /tmp/nextflow.sock
# [ Nextflow #N ]──┘         ▲
#                            │
#              [FastAPI UDS Server]
#                       │
#             +─────────┴─────────+
#             │ asyncio.Queue/msg │
#             │   optional logic  │
#             └─────▶ SSE 推送 ───┘


#  [Producer] --->
#                \
#  [Producer] ---> \                     ┌────────────┐
#                    --> [UDS Server] -->│  asyncio   │
#  [Producer] ---> /                     │   Queue    │--> 处理 / 落盘 / SSE
#                /                       └────────────┘
#  [Producer] --->


# UDSListener
#    ↓
# EventRouter
#    ↓
# WorkflowQueueManager
#    ↓
# PubSubManager
# from brave.api.core.pubsub import PubSubManager
# from .workflow_queue import WorkflowQueueManager
# from brave.api.core.ingress_event_router import IngressEventRouter
# from .sse import SSEManager
# from brave.api.ingress.http_ingress import HTTPIngress
# from dependency_injector.wiring import inject, Provide
# from brave.app_container import AppContainer
# # from .config import EVENT_MODE, UDS_PATH
# EVENT_MODE = "stream"
# UDS_PATH = "/tmp/brave.sock"

# class WorkflowManager:
#     @inject
#     def __init__(
#         self, 
#         queue_manager: WorkflowQueueManager = Provide[AppContainer.workflow_queue_manager], 
#         pubsub: PubSubManager = Provide[AppContainer.pubsub_manager]):

#         self.pubsub = pubsub
#         self.queue_manager = queue_manager
#         self.sse = SSEManager(self.queue_manager)

#     # async def start(self):
#     #     if isinstance(self.ingress, BaseMessageIngress):
#     #         await self.ingress.start()

#     # def register_http(self, app):
#     #     if isinstance(self.ingress, HTTPIngress):
#     #         self.ingress.register(app)

#     def sse_endpoint(self):
#         return self.sse.create_endpoint()

#     def register_subscriber(self, topic: str, callback):
#         self.pubsub.subscribe(topic, callback)

#     async def cleanup_loop(self):
#         while True:
#             await self.queue_manager.cleanup()
#             await asyncio.sleep(10)




# class WorkflowEventSystem:
#     def __init__(self, uds_path="/tmp/brave.sock"):
#         self.pubsub = PubSubManager()
#         self.queue_manager = WorkflowQueueManager(self.pubsub)
#         self.router = EventRouter(self.pubsub, self.queue_manager)
#         self.uds_listener = UDSListener(uds_path, self.router)
#         self.sse = SSEManager(self.queue_manager)

#     async def start(self):
#         await self.uds_listener.start()

#     def sse_endpoint(self):
#         return self.sse.create_endpoint()

#     def register_subscriber(self, topic: str, callback):
#         self.pubsub.subscribe(topic, callback)

# class WorkflowEventSystem:
#     def __init__(self, uds_path="/tmp/brave.sock"):
#         self.pubsub = PubSubManager()
#         self.queue_manager = WorkflowQueueManager(self.pubsub)
#         self.router = EventRouter(self.pubsub, self.queue_manager)
#         self.uds_listener = UDSListener(uds_path, self.router)
#         self.sse = SSEManager(self.queue_manager)

#     async def start(self):
#         await self.uds_listener.start()

#     def sse_endpoint(self):
#         return self.sse.create_endpoint()


from fastapi import APIRouter
from brave.api.config.config import get_settings
setting_controller = APIRouter()

@setting_controller.get("/setting/get-setting")
def get_setting():
    get_setting = get_settings()
    setting_dict = get_setting.__dict__
    return setting_dict


import os
from fastapi import APIRouter
from brave.api.config.config import get_settings
import glob
import json
from collections import defaultdict
from brave.api.config.db import get_engine
import brave.api.service.pipeline as pipeline_service

component_store_api = APIRouter(prefix="/component-store",tags=["component_store"])

def open_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as f:
        data = json.load(f)
    data["file_path"] = file_path
    
    return data

@component_store_api.get("/list")
async def list_components():
    settings = get_settings()
    store_dir = settings.STORE_DIR
    file_list = glob.glob(f"{store_dir}/*/*/*/install.json")
    file_list = [open_file(file) for file in file_list]
    component_ids= [ item["component_id"] for item in file_list]
    with get_engine().begin() as conn: 
        component_installed =  pipeline_service.find_by_component_ids(conn,component_ids)
    
    installed_dict = { item["component_id"]:item for item in component_installed}
    for item in file_list:
        if item["component_id"] in installed_dict:
            item["installed"] = True
        else:
            item["installed"] = False
    file_dict = defaultdict(list)
    
    for item in file_list:
        if "script" in item["file_path"]:
            file_dict["script"].append(item)
        elif "software" in item["file_path"]:
            file_dict["software"].append(item)
        elif "pipeline" in item["file_path"]:
            file_dict["pipeline"].append(item)
        elif "file" in item["file_path"]:
            file_dict["file"].append(item)
        if "img" in item and item["img"] !="":
            img_dir = item["file_path"].replace(str(settings.STORE_DIR),"")
            img_dir = os.path.dirname(img_dir)
            img_name=item["img"]
            item["img"] = f"/brave-api/store-dir{img_dir}/{img_name}"
            
    return file_dict


@component_store_api.get("/list-by-type/{component_type}")
async def list_components_by_type(component_type:str):
    components = await list_components()
    if component_type not in components:
        return []
    return components[component_type]


# 「消费全局队列并广播给每个客户端」 vs 「直接向所有客户端广播」

from fastapi import FastAPI, Request
from fastapi.responses import StreamingResponse
import asyncio
from typing import Set

app = FastAPI()

# 所有活跃连接的队列
clients: Set[asyncio.Queue] = set()

# 发送广播给所有客户端
async def broadcast_message(message: str):
    for queue in clients.copy():
        await queue.put(message)

# 每个连接独立的 SSE 生成器
async def event_generator(request: Request, client_queue: asyncio.Queue):
    try:
        while True:
            # 客户端断开连接则退出
            if await request.is_disconnected():
                break
            try:
                # 等待消息
                message = await asyncio.wait_for(client_queue.get(), timeout=10)
                yield f"data: {message}\n\n"
            except asyncio.TimeoutError:
                # 心跳，防断开
                yield ": keep-alive\n\n"
    finally:
        # 连接断开，清除客户端队列
        clients.discard(client_queue)

@app.get("/sse")
async def sse(request: Request):
    # 为新客户端创建一个队列
    client_queue = asyncio.Queue()
    clients.add(client_queue)
    return StreamingResponse(event_generator(request, client_queue), media_type="text/event-stream")

@app.get("/send")
async def send_msg(msg: str):
    await broadcast_message(msg)
    return {"status": "sent", "message": msg}



# https://github.com/FaztWeb/fastapi-mysql-restapi/blob/main/routes/user.py

from fastapi import FastAPI, Request
from contextlib import asynccontextmanager
from dependency_injector import providers
from fastapi.encoders import jsonable_encoder

from brave.api.routes.file_parse_plot import file_parse_plot
from brave.api.routes.sample_result import sample_result
from brave.api.routes.sample import sample
from brave.api.routes.analysis import analysis_api
from brave.api.routes.pipeline import pipeline
from brave.api.routes.literature import literature_api
from brave.api.routes.sse import sseController
import asyncio
# from brave.api.service.watch_service import watch_folder,startup_process_event
from brave.api.service.sse_service import SSESessionService
from brave.api.service.process_monitor_service import ProcessMonitor
from brave.api.routes.bio_database import bio_database
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse
from fastapi.responses import JSONResponse

import os
from brave.api.config.config import get_settings,init as init_config
from brave.api.service.sse_service import  SSEService  # 从 service.py 导入
from brave.api.routes.namespace import namespace
from brave.api.routes.file_operation import file_operation  
# from brave.api.service.sse_service import get_sse_service
# from brave.api.service.analysis_result_parse import get_analysis_result_parse_service
from brave.api.service.listener_files_service import get_listener_files_service
from brave.api.routes.setting import setting_controller
from brave.app_manager import AppManager
from brave.app_container import AppContainer
from brave.setup_routes import setup_routes
from fastapi.middleware.cors import CORSMiddleware

container = AppContainer()
# container.config.executer_type.from_env("EXECUTER_TYPE","local")    

container.wire(modules=["brave.api.routes"], packages=["brave.api.routes","brave.api.handlers"])
# 

@asynccontextmanager
async def lifespan(app: FastAPI):
    settings = get_settings()
    container.config.from_dict({
        'executer_type': settings.EXECUTER_TYPE
    })

    manager = AppManager()
    # container.app_manager.override(providers.Object(manager))
    await manager.start()
    setup_routes(app,manager)
    init_config(app)
    app.state.manager = manager 
    yield 
    await manager.stop()

# @asynccontextmanager
# async def lifespan(app: FastAPI):
#     settings = get_settings()
#     print("✅ 启动后台任务")
#     current_loop = asyncio.get_event_loop()
#     # print(f"startup 事件循环：{current_loop}")
#     global producer_task, broadcast_task
#     watch_path = f"{settings.BASE_DIR}/monitor"
#     if not  os.path.exists(watch_path):
#         os.makedirs(watch_path)

#     # asyncio.create_task(watch_folder(monitor))
#     # sse_service = SSEService()
#     # sse_service = SSESessionService()
#     sse_service = get_sse_service()
#     listener_files_service = get_listener_files_service()
#     analysis_result_parse_service = get_analysis_result_parse_service(
#         sse_service=sse_service,
#         listener_files_service=listener_files_service
#     )
#     asyncio.create_task(analysis_result_parse_service.auto_save_analysis_result())
#     file_watcher = FileWatcher(
#         watch_path=watch_path,
#         sse_service=sse_service,
#         analysis_result_parse_service=analysis_result_parse_service,
#         listener_files_service=listener_files_service
#     )
#     process_monitor = ProcessMonitor(
#         sse_service=sse_service,
#         analysis_result_parse_service=analysis_result_parse_service,
#         listener_files_service=listener_files_service)

#     app.state.sse_service = sse_service
#     app.state.file_watcher = file_watcher
#     app.state.process_monitor = process_monitor
#     asyncio.create_task(file_watcher.watch_folder())
#     asyncio.create_task(sse_service.broadcast_loop())
#     # asyncio.create_task(sse_service.producer())
#     asyncio.create_task(process_monitor.startup_process_event())

#     wes = WorkflowEventSystem()
#     wes.register_http(app)
#     wes_task = asyncio.create_task(wes.start())
#     # get_analysis_result_parse(sse_service)
#     yield
#     if wes_task: 
#         wes_task.cancel()
#         print("[System] WorkflowEventSystem stopped")
    
def create_app() -> FastAPI:
    app = FastAPI(lifespan=lifespan)
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],      # 放开所有来源
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    # 异常处理函数
    async def value_error_handler(request: Request, exc: Exception):
        return JSONResponse(
            status_code=400,
            content={"detail": str(exc)},
            headers={"Access-Control-Allow-Origin": "*"}
        )

    # 注册异常处理器
    app.add_exception_handler(Exception, value_error_handler)
    # # @app.exception_handler(Exception)
    # async def value_error_handler(request: Request, exc: ValueError):
    #      return JSONResponse(
    #             status_code=400,
    #             content=jsonable_encoder({"detail":"aa", "body":"aa"}),
    #         )
  
    return app

# import logging
# import sys

# logging.basicConfig(
#     level=logging.INFO,
#     handlers=[
#         logging.StreamHandler(sys.stdout),
#         logging.StreamHandler(sys.stderr),
#         logging.FileHandler("brave.log")
#     ]
# )



from fastapi import APIRouter, HTTPException,Query
from brave.api.config.config import  get_graph
from py2neo import Graph, Node, Relationship

from brave.api.config.db import get_engine
from brave.microbe.schemas.entity import GraphQuery, PageEntity, RelationshipRequest,GraphQueryV2
from brave.microbe.service import association_service, graph_service
entity_relation_api = APIRouter(prefix="/entity-relation")

@entity_relation_api.get("/test")
async def test_entity_relation():
    graph = get_graph()
    return {"message": "Entity relation endpoint is working"}



@entity_relation_api.post("/relationships")
async def create_relationship(req: RelationshipRequest):
    graph = get_graph()
    # MERGE 节点 A
    query = f"""
    MERGE (a:{req.from_entity.label} {{entity_id: $from_id}})
    ON CREATE SET a += $from_props
    ON MATCH SET a += $from_props

    MERGE (b:{req.to_entity.label} {{entity_id: $to_id}})
    ON CREATE SET b += $to_props
    ON MATCH SET b += $to_props

    MERGE (a)-[r:{req.relation_type}]->(b)
    RETURN a, r, b
    """
    params = {
        "from_id": req.from_entity.entity_id,
        "from_props": req.from_entity.properties,
        "to_id": req.to_entity.entity_id,
        "to_props": req.to_entity.properties,
    }
    result = graph.run(query, **params).data()
    return {"result": result}


@entity_relation_api.post("/graph-v2")
async def get_graph_relations(graphQuery: GraphQuery):
    """
    获取图数据，包括 Taxonomy、干预、文献和疾病关系
    - label: 过滤节点类型，如 Taxonomy
    - keyword: 过滤节点名称包含关键词
    - entity_id: 指定节点
    """
    graph = get_graph()
    nodes_dict = {}
    links = {}
    graphQuery.collect_association_study = True
    query, params,node_var_map = graph_service.build_cypher(graphQuery)
    query = graph_service.build_return_cypher(query,graphQuery)
    query = graph_service.build_order_cypher(query,graphQuery)
    
    query += " LIMIT 100"
    print(query)
    
    #     WITH a,t,d,i, collect(DISTINCT s.entity_id) AS study_ids,
    #     collect(DISTINCT s.entity_name) AS study_names
    # SET a.study_ids = study_ids,
    #     a.study_names = study_names
            #  collect(DISTINCT {id: s.entity_id, name: s.entity_name}) AS studies 
    # WITH a,t,d,i, collect(DISTINCT {id: s.entity_id, name: s.entity_name}) AS study_list
    # SET a.studies = study_list
            #    s.entity_id AS study_id, s.entity_name AS study_name,


    result = graph.run(query, **params).data()

    for record in result:
        if  record["association_id"] not in nodes_dict:
            entity_name = record["effect"]
            if len(record["studies"]) > 0:
                entity_name += " (" + ", ".join([s["name"] for s in record["studies"]]) + ")"
            nodes_dict[record["association_id"]] = {
                "id": record["association_id"],
                "label": "association",
                "entity_name": entity_name,
                "effect": record["effect"],
                "studies": record["studies"]
            }
        # Taxonomy 节点
        if "taxonomy" in graphQuery.nodes and  record["taxonomy_id"] not in nodes_dict:
            nodes_dict[record["taxonomy_id"]] = {
                "id": record["taxonomy_id"],
                "label": "taxonomy",
                "entity_name": record["taxonomy_name"],
                "taxonomy_links": record["taxonomy_links"]
            }
        # Association 节点
       
        # Disease 节点
        if "disease" in graphQuery.nodes and record["disease_id"] and record["disease_id"] not in nodes_dict:
            nodes_dict[record["disease_id"]] = {
                "id": record["disease_id"],
                "label": "disease",
                "entity_name": record["disease_name"]
            }
        # Study 节点
        if "study" in graphQuery.nodes and  record["study_id"] and record["study_id"] not in nodes_dict:
            nodes_dict[record["study_id"]] = {
                "id": record["study_id"],
                "label": "study",
                "entity_name": record["study_name"]
            }
        # Intervention 节点
        if "diet_and_food" in graphQuery.nodes and record["diet_and_food_id"] and record["diet_and_food_id"] not in nodes_dict:
            nodes_dict[record["diet_and_food_id"]] = {
                "id": record["diet_and_food_id"],
                "label": "diet_and_food",
                "entity_name": record["diet_and_food_name"]
            }

        # 添加关系
        links.setdefault(record["association_id"], set())
        if "diet_and_food" in graphQuery.nodes and  record["diet_and_food_id"]:
            links[record["association_id"]].add(("SUBJECT", record["diet_and_food_id"]))
        if "disease" in graphQuery.nodes and  record["disease_id"]:
            links[record["association_id"]].add(("OBJECT", record["disease_id"]))
        if "study" in graphQuery.nodes and  record["study_id"]:
            links[record["association_id"]].add(("EVIDENCED_BY", record["study_id"]))
        # Taxonomy 节点可以保留原来的 SUBJECT 关系，也可以改成 OBJECT，看业务需求
        if  "taxonomy" in graphQuery.nodes and record["taxonomy_id"]:
            links[record["association_id"]].add(("TAXONOMY", record["taxonomy_id"]))

    # 转换 links 为前端可用列表
    links_list = [{"source": source, "target": target, "type": rel_type}
                  for source, targets in links.items() for rel_type, target in targets]

    return {
        "nodes": list(nodes_dict.values()),
        "links": links_list
    }



@entity_relation_api.post("/graph")
async def get_graph_relations(graphQuery: GraphQuery):
    """
    获取图数据
    - label: 过滤节点类型，如 Study, Disease, Taxonomy
    - keyword: 过滤节点名称包含关键词
    """
    graph = get_graph()
    label = graphQuery.label
    keyword = graphQuery.keyword
    entity_id = graphQuery.entity_id

    # 基础查询
    # query = "MATCH (a)-[r*1..3]->(b) WHERE 1=1 "
    query ="""
    MATCH (a)-[r]->(b)
    WHERE 1=1
    """
    params = {}

    # 按标签过滤
    if label:
        query += "AND ($label IN labels(a) OR $label IN labels(b)) "
        params["label"] = label

    # 按关键词过滤
    if keyword:
        query += "AND (a.entity_name CONTAINS $keyword OR b.entity_name CONTAINS $keyword) "
        params["keyword"] = keyword
    # 按 entity_id 过滤
    if entity_id:
        query += "AND (a.entity_id = $entity_id OR b.entity_id = $entity_id) "
        params["entity_id"] = entity_id

    if graphQuery.locale and graphQuery.locale=="zh_CN":
        query += """
        WITH a, b, collect({association_id:r.association_id,type: type(r), effect: r.effect,study_name:r.study_name,predicate:r.predicate}) AS relations
        RETURN a.entity_id AS from_id, labels(a) AS from_label, 
            coalesce(a.entity_name_zh, a.entity_name) AS from_name,
            a.entity_name AS from_name_en,
            b.entity_id AS to_id, labels(b) AS to_label, 
            coalesce(b.entity_name_zh, b.entity_name) AS to_name,
            b.entity_name AS to_name_en,
            relations
        """
    else:
        query += """
        WITH a, b, collect({association_id:r.association_id,type: type(r), effect: r.effect,study_name:r.study_name,predicate:r.predicate}) AS relations
        RETURN a.entity_id AS from_id, labels(a) AS from_label, a.entity_name AS from_name,
            b.entity_id AS to_id, labels(b) AS to_label, b.entity_name AS to_name,
            relations
        """

    result = graph.run(query, **params).data()

    nodes_dict = {}
    links = []

    for record in result:
        # 添加节点（去重）
        if record["from_id"] not in nodes_dict:
            nodes_dict[record["from_id"]] = {
                "id": record["from_id"],
                "label": record["from_label"][0] if record["from_label"] else "Unknown",
                "entity_name": record["from_name"]
            }
        if record["to_id"] not in nodes_dict:
            nodes_dict[record["to_id"]] = {
                "id": record["to_id"],
                "label": record["to_label"][0] if record["to_label"] else "Unknown",
                "entity_name": record["to_name"]
            }

        # 添加关系
        links.append({
            "source": record["from_id"],
            "target": record["to_id"],
            "relations": record["relations"]
        })

    return {
        "nodes": list(nodes_dict.values()),
        "links": links
    }

@entity_relation_api.post("/graph-v2")
async def get_graph_relations(graphQuery: GraphQuery):
    """
    获取图数据，包括游离节点
    """
    graph = get_graph()
    label = graphQuery.label
    keyword = graphQuery.keyword
    entity_id = graphQuery.entity_id

    params = {}
    where_clauses = []

    if label:
        where_clauses.append("$label IN labels(n)")
        params["label"] = label
    if keyword:
        where_clauses.append("n.entity_name CONTAINS $keyword")
        params["keyword"] = keyword
    if entity_id:
        where_clauses.append("n.entity_id = $entity_id")
        params["entity_id"] = entity_id

    # 1️ 查询节点
    node_query = "MATCH (n) "
    if where_clauses:
        node_query += "WHERE " + " AND ".join(where_clauses) + " "
    node_query += "RETURN n.entity_id AS id, labels(n) AS labels, n.entity_name AS entity_name, id(n) AS node_id"

    node_results = graph.run(node_query, **params).data()

    nodes_dict = {}
    for node in node_results:
        nodes_dict[node["id"]] = {
            "id": node["id"],
            "node_id": node["node_id"],
            "label": node["labels"][0] if node["labels"] else "Unknown",
            "entity_name": node["entity_name"]
        }

    # 2️ 查询关系
    rel_query = "MATCH (a)-[r]->(b) WHERE 1=1 "
    if label:
        rel_query += "AND ($label IN labels(a) OR $label IN labels(b)) "
    if keyword:
        rel_query += "AND (a.entity_name CONTAINS $keyword OR b.entity_name CONTAINS $keyword) "
    if entity_id:
        rel_query += "AND (a.entity_id = $entity_id OR b.entity_id = $entity_id) "

    rel_query += """
    RETURN a.entity_id AS from_id, a.labels AS from_labels, a.entity_name AS from_name, id(a) AS from_node_id,
           b.entity_id AS to_id, b.labels AS to_labels, b.entity_name AS to_name, id(b) AS to_node_id,
           type(r) AS relation_type, id(r) AS relation_id
    LIMIT 100
    """

    rel_results = graph.run(rel_query, **params).data()

    links = []
    for rel in rel_results:
        # 将关系两端节点加入 nodes_dict
        if rel["from_id"] not in nodes_dict:
            nodes_dict[rel["from_id"]] = {
                "id": rel["from_id"],
                "node_id": rel["from_node_id"],
                "label": rel["from_labels"][0] if rel["from_labels"] else "Unknown",
                "entity_name": rel["from_name"]
            }
        if rel["to_id"] not in nodes_dict:
            nodes_dict[rel["to_id"]] = {
                "id": rel["to_id"],
                "node_id": rel["to_node_id"],
                "label": rel["to_labels"][0] if rel["to_labels"] else "Unknown",
                "entity_name": rel["to_name"]
            }

        # 添加关系
        links.append({
            "relation_id": rel["relation_id"],
            "source": rel["from_id"],
            "target": rel["to_id"],
            "type": rel["relation_type"]
        })

    return {
        "nodes": list(nodes_dict.values()),
        "links": links
    }

@entity_relation_api.get("/relation/{relation_id}")
async def get_relation_by_id(relation_id: int):
    graph = get_graph()

    query = """
    MATCH (a)-[r]->(b)
    WHERE id(r) = $rid
    RETURN id(r) AS rid, type(r) AS type,
           a.entity_id AS from_id, labels(a) AS from_labels, a.entity_name AS from_name,
           b.entity_id AS to_id, labels(b) AS to_labels, b.entity_name AS to_name,
           properties(r) AS props
    """
    params = {"rid": relation_id}
    record = graph.run(query, **params).data()

    if not record:
        raise HTTPException(status_code=404, detail=f"Relation {relation_id} not found")

    # 如果用 .evaluate 返回的是 dict，则直接返回
    return record[0] 


@entity_relation_api.delete("/relation/{relation_id}")
async def delete_relation_by_id(relation_id: int):
    graph = get_graph()

    # # 先获取关系两端节点的 id
    # query_nodes = """
    # MATCH (a)-[r]->(b)
    # WHERE id(r) = $rid
    # RETURN id(a) AS from_id, id(b) AS to_id
    # """
    # record = graph.run(query_nodes, rid=relation_id).data()
    # if not record:
    #     raise HTTPException(status_code=404, detail=f"Relation {relation_id} not found")

    # from_id = record[0]["from_id"]
    # to_id = record[0]["to_id"]

    # 删除关系
    query_delete_relation = """
    MATCH ()-[r]->()
    WHERE id(r) = $rid
    DELETE r
    """
    graph.run(query_delete_relation, rid=relation_id)

    # # 删除孤立节点（没有任何关系的节点）
    # query_delete_isolated_nodes = """
    # MATCH (n)
    # WHERE id(n) IN [$from_id, $to_id] AND NOT (n)--()
    # DELETE n
    # """
    # graph.run(query_delete_isolated_nodes, from_id=from_id, to_id=to_id)

    # return {"message": f"Relation {relation_id} deleted successfully, isolated nodes removed if any."}
    return {"message": f"Relation {relation_id} deleted successfully."}



@entity_relation_api.get("/find-by-paired-entity/{from_entity}/{to_entity}")
async def find_by_paired_entity(from_entity: str, to_entity: str):
    """
    根据两个实体 ID 查找它们之间的所有关系
    """
    graph = get_graph()
    
    # 使用 Cypher 查询
    query = """
    MATCH (a)-[r]->(b)
    WHERE a.entity_id = $from_entity AND b.entity_id = $to_entity
    RETURN type(r) AS relation_type, r, a.entity_id AS from_id, b.entity_id AS to_id
    """
    
    result = graph.run(query, from_entity=from_entity, to_entity=to_entity)
    
    # 返回列表，每条关系包括类型和属性
    relationships = []
    for record in result:
        relationships.append({
            "relation_type": record["relation_type"],
            "properties": dict(record["r"]),
            "from_id": record["from_id"],
            "to_id": record["to_id"]
        })
    
    return relationships

def delete_relationship_by_association_id(association_id: str):
    graph = get_graph()
    query = """
    MATCH ()-[r]->()
    WHERE r.association_id = $association_id
    DELETE r
    """
    graph.run(query, association_id=association_id)
    # print(f"Deleted relationship with association_id={association_id}")

@entity_relation_api.delete("/delete-association/{asso_id}")
async def delete_entity( asso_id: str): 
    delete_relationship_by_association_id(asso_id)
    with get_engine().begin() as conn:

        association_service.delete_by_id(conn, asso_id)

    return {"message": f"删除成功 {asso_id}"}

@entity_relation_api.post("/page")
async def page_entity( query: PageEntity):
    with get_engine().begin() as conn:
        result = association_service.page(conn, query)
    return result


import glob
import re
from functools import reduce

from brave.api.config.config import get_settings

def glob_to_regex(glob_path: str) -> str:
    # 转义路径中的正则特殊字符（除了 *）
    escaped = re.escape(glob_path)
    # 把转义的 \* 替换成 (.+)
    regex = escaped.replace(r'\*', r'(.+)')
    return f'r"{regex}"'

def from_glob_get_file(content,dir=None):
    settings = get_settings()
    form_data = {}
    for k,v in content.items():
        if dir: 
            pattern_str = f"{dir}/{v}".strip()
        else:
            pattern_str = v.strip()

        if pattern_str.startswith("~"):
            pattern_str = pattern_str.replace("~",str(settings.DATA_DIR))

        
        file_list = glob.glob(pattern_str)
        pattern = re.compile(glob_to_regex(pattern_str)[2:-1]) 
        result_dict = {}
        for file in file_list:
            match = pattern.match(file)
            if match:
                # match_dict = match.groupdict()
                file_name = match.group(1)
                result_dict[file_name] = file
        form_data[k] = result_dict
    # common_samples = reduce(lambda  x,y: set(x.keys()) & set(y.keys()), list(form_data.values()))
    # dicts = list(form_data.values())
    # common_samples = reduce(lambda x, y: set(x) & set(y), (d.keys() for d in dicts))
    common_samples =  set.intersection(*(set(d.keys()) for d in form_data.values()))
    result = []
    for name in  common_samples:
        result_dict = { "sample_name":name}
        for k,files in  form_data.items():
            result_dict.update({
                k:files[name]
            })
        result.append(result_dict)

    return result


from enum import Enum
from typing import Any, Type, TypeVar

from pydantic import BaseModel
from brave.api.core.base_event_router import BaseEventRouter
from brave.api.core.routers_name import RoutersName

# T = TypeVar("T", bound=Any)

class EventBus:
    def __init__(self):
        self.routers: dict[RoutersName, BaseEventRouter] = {}

    def register_router(self, name: RoutersName, router: BaseEventRouter):
        self.routers[name] = router

    async def dispatch(self, router_name: RoutersName, event_type:Enum, *payload: BaseModel):
        router = self.routers.get(router_name)
        if router:
            await router.dispatch(event_type, *payload)
        else:
            print(f"[EventBus] No router found for '{router_name}'")



# brave/workflow_event_system.py

import asyncio
import time
import json
import socket
import os
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable, Coroutine, Union, Literal
from fastapi import Request
from starlette.responses import StreamingResponse

# Type alias
Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine]]

# ================================
# PubSubManager
# ================================
class PubSubManager:
    def __init__(self):
        self.subscribers: dict[str, set[Callback]] = defaultdict(set)

    def subscribe(self, topic: str, callback: Callback):
        self.subscribers[topic].add(callback)

    def unsubscribe(self, topic: str, callback: Callback):
        self.subscribers[topic].discard(callback)

    async def publish(self, topic: str, message: dict):
        for cb in list(self.subscribers.get(topic, [])):
            try:
                if asyncio.iscoroutinefunction(cb):
                    await cb(message)
                else:
                    cb(message)
            except Exception as e:
                print(f"[PubSub ERROR] {e}")


# ================================
# WorkflowQueueManager
# ================================
@dataclass
class WorkflowQueue:
    queue: asyncio.Queue
    task: asyncio.Task
    last_active: float = field(default_factory=time.time)
    subscribers: int = 0

class WorkflowQueueManager:
    def __init__(self, pubsub: PubSubManager):
        self.workflow_map: dict[str, WorkflowQueue] = {}
        self.pubsub = pubsub

    def register(self, workflow_id: str):
        if workflow_id not in self.workflow_map:
            queue = asyncio.Queue()
            task = asyncio.create_task(self._consume_loop(workflow_id, queue))
            self.workflow_map[workflow_id] = WorkflowQueue(queue=queue, task=task)

    async def put(self, workflow_id: str, msg: dict):
        self.register(workflow_id)
        wfq = self.workflow_map[workflow_id]
        wfq.last_active = time.time()
        await wfq.queue.put(msg)

    def get(self, workflow_id: str) -> WorkflowQueue:
        return self.workflow_map[workflow_id]

    async def _consume_loop(self, workflow_id: str, queue: asyncio.Queue):
        while True:
            msg = await queue.get()
            try:
                await self.pubsub.publish(workflow_id, msg)
            except Exception as e:
                print(f"[Consumer ERROR] workflow {workflow_id}: {e}")

    async def cleanup(self, timeout: int = 300):
        now = time.time()
        to_delete = []
        for wf_id, wfq in self.workflow_map.items():
            if wfq.subscribers == 0 and wfq.queue.empty() and (now - wfq.last_active > timeout):
                wfq.task.cancel()
                to_delete.append(wf_id)
        for wf_id in to_delete:
            del self.workflow_map[wf_id]
            print(f"[Cleanup] Removed workflow {wf_id}")


# ================================
# SSEManager
# ================================
class SSEManager:
    def __init__(self, wq_manager: WorkflowQueueManager):
        self.wq_manager = wq_manager

    def create_endpoint(self):
        async def sse_endpoint(workflow_id: str, request: Request):
            self.wq_manager.register(workflow_id)
            wfq = self.wq_manager.get(workflow_id)
            wfq.subscribers += 1
            wfq.last_active = time.time()

            async def event_stream():
                try:
                    while True:
                        if await request.is_disconnected():
                            break
                        msg = await wfq.queue.get()
                        yield f"data: {json.dumps(msg)}\n\n"
                finally:
                    wfq.subscribers -= 1
                    wfq.last_active = time.time()

            return StreamingResponse(event_stream(), media_type="text/event-stream")

        return sse_endpoint


# ================================
# UDSListener using asyncio.start_unix_server
# ================================
class UDSListener:
    def __init__(self, uds_path: str, wq_manager: WorkflowQueueManager):
        self.uds_path = uds_path
        self.wq_manager = wq_manager

    async def start(self):
        if os.path.exists(self.uds_path):
            os.remove(self.uds_path)

        server = await asyncio.start_unix_server(self.handle_client, path=self.uds_path)
        print(f"[UDS] Listening at {self.uds_path}")
        async with server:
            await server.serve_forever()

    async def handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        try:
            while line := await reader.readline():
                try:
                    msg = json.loads(line.decode().strip())
                    workflow_id = msg.get("workflow_id", "global")
                    await self.wq_manager.put(workflow_id, msg)
                except Exception as e:
                    print(f"[UDS] Message error: {e}")
        except Exception as e:
            print(f"[UDS] Client error: {e}")
        finally:
            writer.close()
            await writer.wait_closed()
